// ============================================================================
//
// Copyright (c) 2001  Lutz Kettner, see the Copyright file for details
//
// ----------------------------------------------------------------------------
//
// file          : Point_order_types_2.h
// revision      : $Revision: 1.1 $
// revision_date : $Date: 2001/09/03 06:13:35 $
// author(s)     : Lutz Kettner
//
// File reader for 2D point order types up to 10 points based on the
// data base by Oswin Aichholzer.
// ============================================================================


#ifndef PTW_POINT_ORDER_TYPES_2_H
#define PTW_POINT_ORDER_TYPES_2_H 1

#include <CGAL/basic.h>
#include <fstream>


namespace PTW {

const int MAX_POINT_ORDER_TYPES_2 = 10;


// Reads point configurations from a data base. Writes each configuration
// as a stream of points to the forward iterator.

template <class Forward_iterator_, class Traits_>
class Point_order_types_2 {
    typedef Forward_iterator_                   Forward_iterator;
    typedef Traits_                             Traits;
    typedef typename Traits::Point_2            Point_2;
    typedef typename Traits::Construct_point_2  Construct_point_2;

    std::ifstream     m_in;     // input file
    Forward_iterator  m_out;    // output area where points are written to
    int               m_n;      // number of points per configuration
    Traits            m_traits; // geometric traits
    unsigned char     buffer[ 4 * MAX_POINT_ORDER_TYPES_2]; // block read
public:
    Point_order_types_2( Traits traits = Traits()) : m_n(0), m_traits(traits){}
    Point_order_types_2( const char* filename, // file to open
                         int n,                // points per configuration
			 Forward_iterator out, // place to write points to
			 Traits traits = Traits())
	: m_in( filename), m_out( out), m_n(n), m_traits( traits) 
    {
	CGAL_precondition( n >= 3 && n <= MAX_POINT_ORDER_TYPES_2);
	if ( ! m_in)
	    n = 0;
    }
    bool good()  const { return bool( m_in); }
    bool error() const { return ! m_in && ! m_in.eof(); }
    bool eof()   const { return m_in.eof(); }

    bool read() {
	// Returns true if read did succeed, false otherwise.
	// Distinguish 8 bit and 16 bit reads
	Construct_point_2 make_point = m_traits.construct_point_2_object();
	Forward_iterator out = m_out;
	int n = m_n;
	if ( n < 9) { // 8 bit point coordinates
	    m_in.read( buffer, 2 * n);
	    for ( int k = 0; k != n; ++k)
		* out++ = make_point( CGAL::UInteger16( buffer[ 2*k]),
				      CGAL::UInteger16( buffer[ 2*k+1]));
	} else { // 16 bit point coordinates
	    m_in.read( buffer, 4 * n);
	    for ( int k = 0; k != n; ++k) {
		unsigned char* p = buffer + 4*k;
		// swap byte order on big endian machines
		#ifdef CGAL_BIG_ENDIAN 
		std::swap( p[0], p[1]);
		std::swap( p[2], p[3]);
		#endif
		CGAL::UInteger16* q = reinterpret_cast<CGAL::UInteger16*>(p);
		* out++ = make_point( q[0], q[1]);
	    }
	}
	return good();
    }
    bool seek( long n) {
        // Positions file stream such that next read() accesses point
        // configuration n.
	// Returns true if read did succeed, false otherwise.
	// Distinguish 8 bit and 16 bit reads.
	if ( m_n < 9) { // 8 bit point coordinates
	    m_in.seekg( n * m_n * 2);
	} else { // 16 bit point coordinates
	    m_in.seekg( n * m_n * 4);
	}
	return good();	
    }
};

} // namespace PTW
#endif // PTW_POINT_ORDER_TYPES_2_H //
// EOF //

