// ============================================================================
//
// Copyright (c) 2001  Lutz Kettner, see the Copyright file for details
//
// ----------------------------------------------------------------------------
//
// file          : PseudoT_2.h
// revision      : $Revision: 1.3 $
// revision_date : $Date: 2001/09/03 06:37:02 $
// author(s)     : Lutz Kettner
//
// Pseudo Triangulation Data Structure and Algorithms
// ============================================================================

#ifndef PTW_PSEUDOT_2_H
#define PTW_PSEUDOT_2_H 1

#include <CGAL/basic.h>
#include <CGAL/Polyhedron_iterator_3.h>
#include <CGAL/Iterator_project.h>
#include <CGAL/function_objects.h>
#include <CGAL/N_step_adaptor_derived.h>
#include <CGAL/HalfedgeDS_default.h>
#include <CGAL/HalfedgeDS_const_decorator.h>
#include <CGAL/HalfedgeDS_decorator.h>
#include <CGAL/Modifier_base.h>
#include <CGAL/IO/Verbose_ostream.h>
#include <PTW/PseudoT_items_2.h>
#include <cstddef>
#include <list>

#ifdef CGAL_REP_CLASS_DEFINED
#include <PTW/PseudoT_default_traits_2.h>
#endif // CGAL_REP_CLASS_DEFINED


namespace PTW {


struct Point_accessor {
    template <class Vertex>
    typename std::iterator_traits<Vertex>::value_type::Point 
    operator()( const Vertex& v) {
	return v->point();
    }
};

struct Point_accessor_deref {
    template <class Vertex>
    typename std::iterator_traits< 
        typename std::iterator_traits<Vertex>::value_type >::value_type::Point 
    operator()( const Vertex& v) {
	return (*v)->point();
    }
};

// template <class Input_iterator, class DA>



template <class Vertex_base>
class I_PseudoT_vertex  : public Vertex_base  {
public:
    typedef Vertex_base                            Base;
    //typedef typename Base::HalfedgeDS              HDS;
    typedef typename Base::Point                   Point;
    typedef typename Base::Point                   Point_2;

    // Handles have to explicitly repeated, although they are derived
    typedef typename Base::Vertex_handle           Vertex_handle;
    typedef typename Base::Halfedge_handle         Halfedge_handle;
    typedef typename Base::Face_handle             Face_handle;
    typedef typename Base::Vertex_const_handle     Vertex_const_handle;
    typedef typename Base::Halfedge_const_handle   Halfedge_const_handle;
    typedef typename Base::Face_const_handle       Face_const_handle;
    typedef typename Base::Halfedge                Halfedge;
    typedef typename Base::Face                    Face;

    // Supported options by HDS.
    typedef typename Base::Supports_vertex_halfedge
                                                  Supports_vertex_halfedge;

    // Circulator category.
    typedef typename Halfedge::Supports_halfedge_prev  Supports_prev;

public:
    // Circulator category.
    typedef CGAL::Polyhedron_circulator_traits<Supports_prev> Ctr;
    typedef typename Ctr::iterator_category circulator_category;

    // Circulators around a vertex and around a facet.
    typedef CGAL::I_Polyhedron_facet_circ<Halfedge_handle, circulator_category>
                                         Halfedge_around_facet_circulator;

    typedef CGAL::I_Polyhedron_vertex_circ< Halfedge_handle, 
                                            circulator_category>
                                        Halfedge_around_vertex_circulator;

    typedef CGAL::I_Polyhedron_facet_circ<
        Halfedge_const_handle,
        circulator_category>       Halfedge_around_facet_const_circulator;

    typedef CGAL::I_Polyhedron_vertex_circ<
        Halfedge_const_handle,
        circulator_category>      Halfedge_around_vertex_const_circulator;



    typedef typename Halfedge_around_vertex_circulator::size_type
        size_type;
    typedef typename Halfedge_around_vertex_circulator::difference_type
        difference_type;

public:
    // We need to repeat the constructors here.
    I_PseudoT_vertex() {}
    I_PseudoT_vertex( const Vertex_base& b) : Vertex_base(b) {}
    I_PseudoT_vertex( const Point& p) : Vertex_base(p) {}

// New Access Functions (not provided in Vertex_base).

    Halfedge_around_vertex_circulator vertex_begin() {
        // a circulator of halfedges around the vertex (clockwise).
        return Halfedge_around_vertex_circulator( halfedge());
    }
    Halfedge_around_vertex_const_circulator vertex_begin() const {
        // a circulator of halfedges around the vertex (clockwise).
        return Halfedge_around_vertex_const_circulator( halfedge());
    }

    size_type degree() const {
	if ( halfedge() == Halfedge_const_handle())
	    return 0;
        return CGAL::circulator_size( vertex_begin());
    }
    size_type marked_degree() const {
	if ( halfedge() == Halfedge_const_handle())
	    return 0;
	Halfedge_const_handle h = halfedge();
	Halfedge_const_handle start = h;
	size_type count = 0;
	do {
	    if ( h->is_marked())
		++count;
	    h = h->next()->opposite();
	} while ( h != start);
	CGAL_postcondition( count <= degree());
        return count;
    }
    bool is_reflex() const {
	// test inner angle at all corners around this vertex 
	// for one angle > pi.
	if ( halfedge() != Halfedge_const_handle()) {
	    CGAL_assertion( &* (halfedge()->vertex()) == this);
	    return halfedge()->is_reflex();
	}
	return false;
    }

    // No longer hidden. Now the restricted version with precondition.
    // sets incident halfedge to h. Precondition: h is incident, i.e.,
    // h->vertex() == v.
    void  set_halfedge( Halfedge_handle hh) {
        CGAL_assertion( &*(hh->vertex()) == this);
        Base::set_halfedge(hh);
    }
};


// A halfedge is an oriented edge. Both orientations exist, i.e.
// an edge is represented by two opposite halfedges. The geometric
// relations are as follows:
//
//              _ _ _   .
//             /        |\.
//                      | \.
//           /             \ next half
//                          \ edge
//         /                 \.
//
//        |                   O  incident vertex
//                facet      ,
//        |                 /| |
//                         / | | opposite
//         \                 | | half edge
//                      half | |
//           \          edge | | /
//                           | |/
//             \_ _ _ _ _ _    '
//

template <class Halfedge_base>
class I_PseudoT_halfedge : public Halfedge_base {
public:
    typedef Halfedge_base                          Base;
    typedef typename Base::HalfedgeDS              HDS;

    typedef typename Base::Traits                  Traits;
    typedef typename Traits::Leftturn_2            Leftturn_2;

    // Handles have to explicitly repeated, although they are derived
    typedef typename Base::Vertex_handle           Vertex_handle;
    typedef typename Base::Halfedge_handle         Halfedge_handle;
    typedef typename Base::Face_handle             Face_handle;
    typedef typename Base::Vertex_const_handle     Vertex_const_handle;
    typedef typename Base::Halfedge_const_handle   Halfedge_const_handle;
    typedef typename Base::Face_const_handle       Face_const_handle;

    typedef typename Base::Vertex                  Vertex;
    typedef typename Base::Face                    Face;

    // Supported options by HDS.
    typedef typename Base::Supports_halfedge_prev Supports_halfedge_prev;
    typedef typename Base::Supports_halfedge_vertex
                                                  Supports_halfedge_vertex;
    typedef typename Base::Supports_halfedge_face Supports_halfedge_face;

    // Circulator category.
    typedef typename Base::Supports_halfedge_prev Supports_prev;

public:
    // Circulator category.
    typedef CGAL::Polyhedron_circulator_traits<Supports_prev> Ctr;
    typedef typename Ctr::iterator_category circulator_category;

    // Circulators around a vertex and around a facet.
    typedef CGAL::I_Polyhedron_facet_circ< Halfedge_handle,circulator_category>
                                         Halfedge_around_facet_circulator;

    typedef CGAL::I_Polyhedron_vertex_circ<Halfedge_handle,circulator_category>
                                        Halfedge_around_vertex_circulator;

    typedef CGAL::I_Polyhedron_facet_circ<
        Halfedge_const_handle,
        circulator_category>       Halfedge_around_facet_const_circulator;

    typedef CGAL::I_Polyhedron_vertex_circ<
        Halfedge_const_handle,
        circulator_category>      Halfedge_around_vertex_const_circulator;

    typedef typename Halfedge_around_vertex_circulator::size_type
        size_type;
    typedef typename Halfedge_around_vertex_circulator::difference_type
        difference_type;

private:
    bool m_mark;

public:
    I_PseudoT_halfedge() : m_mark(false) {}
    I_PseudoT_halfedge( const Halfedge_base& b)
	: Halfedge_base(b), m_mark(false) {}

// New Access Functions (not provided in Halfedge_base).

    // Change semantic of prev: it is always available at this level.
    // If the HDS does not provide a prev-function, the previous
    // halfedge will be searched around the incident facet.
    Halfedge_handle       prev() {
        CGAL::HalfedgeDS_items_decorator<HDS> decorator;
        return decorator.find_prev( HDS::halfedge_handle(this));
    }
    Halfedge_const_handle prev() const {
        CGAL::HalfedgeDS_items_decorator<HDS> decorator;
        return decorator.find_prev( HDS::halfedge_handle(this));
    }

    // the next halfedge around the vertex (clockwise). This is equal to
    // `h.next()->opposite()'.
    Halfedge_handle       next_on_vertex() { return next()->opposite(); }
    Halfedge_const_handle next_on_vertex() const {
        return next()->opposite();
    }

    // the previous halfedge around the vertex (counterclockwise). Is
    // equal to `h.opposite()->prev()'.
    Halfedge_handle       prev_on_vertex() { return opposite()->prev(); }
    Halfedge_const_handle prev_on_vertex() const {
        return opposite()->prev();
    }

    // a circulator of halfedges around the vertex (clockwise).
    Halfedge_around_vertex_circulator vertex_begin() {
        return Halfedge_around_vertex_circulator(
            HDS::halfedge_handle(this));
    }
    Halfedge_around_vertex_const_circulator vertex_begin() const {
        return Halfedge_around_vertex_const_circulator(
            HDS::halfedge_handle(this));
    }

    // a circulator of halfedges around the facet (counterclockwise).
    Halfedge_around_facet_circulator facet_begin() {
        return Halfedge_around_facet_circulator(
            HDS::halfedge_handle(this));
    }
    Halfedge_around_facet_const_circulator facet_begin() const {
        return Halfedge_around_facet_const_circulator(
            HDS::halfedge_handle(this));
    }
    bool is_triangle() const {
	return this != &* ( next()) && this == &* ( next()->next()->next());
    }
    bool is_corner() const {
	// test inner angle at this vertex < pi.
	Leftturn_2 leftturn;
	return leftturn( opposite()->vertex()->point(), vertex()->point(),
			 next()->vertex()->point());
    }
    bool is_reflex() const {
	// test inner angle at all corners around this vertex 
	// for one angle > pi.
	Leftturn_2 leftturn; // Hack: I don't want a traits object in 
                             // each halfedge
	Halfedge_const_handle h = next_on_vertex();
  	Halfedge_const_handle start = h;
  	do {
  	    if ( leftturn( h->next()->vertex()->point(),
			   h->vertex()->point(),
  			   h->opposite()->vertex()->point()))
  		return true;
	    h = h->next()->opposite();
  	} while ( h != start);
  	return false;
    }
    size_type size_of_corners() const {
	size_type count = 0;
	Halfedge_const_handle h = next();
	Halfedge_const_handle start = h;
	do {
	    if ( h->is_corner())
		++ count;
	    h = h->next();
	} while ( h != start);
	return count;
    }
    bool is_pseudo_triangle() const {
	return size_of_corners() == 3;
    }

    bool get_corners( Halfedge_const_handle& h1, 
		      Halfedge_const_handle& h2, 
		      Halfedge_const_handle& h3) const {
	Halfedge_const_handle h = next();
	Halfedge_const_handle start = h;
	do {
	    if ( h->is_corner()) {
		h1 = h;
		h = h->next();
		break;
	    }
	    h = h->next();
	    if ( h == start)
		return false;
	} while ( true);
	do {
	    if ( h->is_corner()) {
		h2 = h;
		h = h->next();
		break;
	    }
	    h = h->next();
	    if ( h == start)
		return false;
	} while ( true);
	do {
	    if ( h->is_corner()) {
		h3 = h;
		h = h->next();
		break;
	    }
	    h = h->next();
	    if ( h == start)
		return false;
	} while ( true);
	return true;
    }
    bool get_corners( Halfedge_handle& h1, 
		      Halfedge_handle& h2, 
		      Halfedge_handle& h3) {
	return ((const I_PseudoT_halfedge<Halfedge_base> *)this)
	    ->get_corners( (Halfedge_const_handle&)h1,
			   (Halfedge_const_handle&)h2,
			   (Halfedge_const_handle&)h3);
    }
    bool is_marked()         const { return m_mark; }
    void set_marked( bool b)       { m_mark = b; }

    double angle() const {
	double phi = M_PI_2 + atan2( 
	    (vertex()->point() - opposite()->vertex()->point()).y(),
	    (vertex()->point() - opposite()->vertex()->point()).x());
	if ( phi <= 0.0 )
	    phi += ( 2 * M_PI);
	return phi;
    }

private:
    // Hide some other functions of H.
    void  set_next( Halfedge_handle hh)  { Base::set_next(hh);}
    void  set_prev( Halfedge_handle hh)  { Base::set_prev(hh);}
    void  set_vertex( Vertex_handle vv)  { Base::set_vertex(vv);}
    void  set_face( Face_handle ff)      { Base::set_face(ff);}
};


template <class Face_base>
class I_PseudoT_face  : public Face_base  {
public:
    typedef Face_base                             Base;
    //typedef typename Base::HalfedgeDS              HDS;

    // Handles have to explicitly repeated, although they are derived
    typedef typename Base::Vertex_handle           Vertex_handle;
    typedef typename Base::Halfedge_handle         Halfedge_handle;
    typedef typename Base::Face_handle             Face_handle;
    typedef typename Base::Vertex_const_handle     Vertex_const_handle;
    typedef typename Base::Halfedge_const_handle   Halfedge_const_handle;
    typedef typename Base::Face_const_handle       Face_const_handle;
    typedef typename Base::Vertex                  Vertex;
    typedef typename Base::Halfedge                Halfedge;

    // Supported options by HDS.
    typedef typename Base::Supports_face_halfedge Supports_face_halfedge;

    // Circulator category.
    typedef typename Halfedge::Supports_halfedge_prev  Supports_prev;

public:
    // Circulator category.
    typedef CGAL::Polyhedron_circulator_traits<Supports_prev> Ctr;
    typedef typename Ctr::iterator_category circulator_category;

    // Circulators around a vertex and around a facet.
    typedef CGAL::I_Polyhedron_facet_circ<Halfedge_handle,circulator_category>
                                         Halfedge_around_facet_circulator;

    typedef CGAL::I_Polyhedron_vertex_circ<Halfedge_handle,circulator_category>
                                        Halfedge_around_vertex_circulator;

    typedef CGAL::I_Polyhedron_facet_circ<
        Halfedge_const_handle,
        circulator_category>       Halfedge_around_facet_const_circulator;

    typedef CGAL::I_Polyhedron_vertex_circ<
        Halfedge_const_handle,
        circulator_category>      Halfedge_around_vertex_const_circulator;

    typedef typename Halfedge_around_vertex_circulator::size_type
        size_type;
    typedef typename Halfedge_around_vertex_circulator::difference_type
        difference_type;

public:
    // We need to repeat the constructors here.
    I_PseudoT_face() {}
    I_PseudoT_face( const Face_base& b) : Face_base(b) {}

// New Access Functions (not provided in Face_base).

    Halfedge_around_facet_circulator face_begin() {
        // a circulator of halfedges around the facet (counterclockwise).
        return Halfedge_around_face_circulator( halfedge());
    }
    Halfedge_around_facet_const_circulator face_begin() const {
        // a circulator of halfedges around the facet (counterclockwise).
        return Halfedge_around_face_const_circulator( halfedge());
    }

    size_type size() const {
        return CGAL::circulator_size( face_begin());
    }
    size_type size_of_corners() const {
	if ( halfedge() != Halfedge_const_handle())
	    return halfedge()->size_of_corners();
	return 0;
    }
    bool is_pseudo_triangle() const {
	return size_of_corners() == 3;
    }

    bool get_corners( Halfedge_const_handle& h1, 
		      Halfedge_const_handle& h2, 
		      Halfedge_const_handle& h3) const {
	if ( halfedge() == Halfedge_const_handle())
	    return false;
	return halfedge()->get_corners( h1, h2, h3);
    }
    bool get_corners( Halfedge_handle& h1, 
		      Halfedge_handle& h2, 
		      Halfedge_handle& h3) {
	if ( halfedge() == Halfedge_handle())
	    return false;
	return halfedge()->get_corners( h1, h2, h3);
    }

    // No longer hidden. Now the restricted version with precondition.
    // sets incident halfedge to h. Precondition: h is incident, i.e.,
    // h->face() == v.
    void  set_halfedge( Halfedge_handle hh) {
        CGAL_assertion( &*(hh->facet()) == this);
        Base::set_halfedge(hh);
    }
};

template < class Items>
class I_PseudoT_derived_items {
public:
    template < class Refs, class Traits>
    class Vertex_wrapper {
    public:
        typedef typename Items::template Vertex_wrapper<Refs,Traits> VWrap;
        typedef typename VWrap::Vertex Vertex_base;
        typedef I_PseudoT_vertex< Vertex_base> Vertex;
    };
    template < class Refs, class Traits>
    class Halfedge_wrapper {
    public:
        typedef typename Items::template Halfedge_wrapper<Refs,Traits> HWrap;
        typedef typename HWrap::Halfedge Halfedge_base;
        typedef I_PseudoT_halfedge< Halfedge_base> Halfedge;
    };
    template < class Refs, class Traits>
    class Face_wrapper {
    public:
        typedef typename Items::template Face_wrapper<Refs,Traits> FWrap;
        typedef typename FWrap::Face Face_base;
        typedef I_PseudoT_face< Face_base> Face;
    };
};


// -------------------------------------------------------------------------
// Main class
// -------------------------------------------------------------------------

template < class p_Traits,
           class p_Items = PseudoT_items_2,
#ifndef CGAL_CFG_NO_TMPL_IN_TMPL_PARAM
           template < class T, class I, class A>
#endif
           class p_HDS = CGAL::HalfedgeDS_default,
           class Alloc = CGAL_ALLOCATOR(int)>
class PseudoT_2 {
public:
    typedef PseudoT_2<p_Traits,p_Items,p_HDS,Alloc>  Self;
    typedef p_Traits                              Traits;
    typedef p_Items                               Items;
    typedef I_PseudoT_derived_items<Items>        Derived_items;
#ifndef CGAL_CFG_NO_TMPL_IN_TMPL_PARAM
    typedef p_HDS< Traits, Derived_items, Alloc>  HDS;
#else
    typedef typename p_HDS::template HDS< Traits, Derived_items, Alloc>  HDS;
#endif
    typedef HDS                                   HalfedgeDS;

    // portability with older CGAL release
    typedef HDS                                   Halfedge_data_structure;

    typedef Alloc                                 Allocator;
    typedef Alloc                                 allocator_type; // STL name

    // Container stuff.
    typedef typename HDS::size_type               size_type;
    typedef typename HDS::difference_type         difference_type;
    typedef typename HDS::iterator_category       iterator_category;
    typedef typename HDS::Supports_removal        Supports_removal;

    // Geometry
    typedef typename Traits::Point_2              Point_2;
    typedef typename Traits::Point_2              Point;

    typedef typename Traits::Construct_point_2    Construct_point_2;
    typedef typename Traits::Less_xy_2            Less_xy_2;
    typedef typename Traits::Leftturn_2           Leftturn_2;
    typedef typename Traits::In_circle_2          In_circle_2;
    typedef typename Traits::Angle_2              Angle_2;
    typedef typename Traits::Rotate_ccw_less_2    Rotate_ccw_less_2;
    typedef typename Traits::Less_signed_distance_to_line_2
                                                Less_signed_distance_to_line_2;

    // Items
    typedef typename HDS::Vertex                  Vertex;
    typedef typename HDS::Halfedge                Halfedge;
    typedef typename HDS::Face                    Face;

    typedef typename Vertex::Base                 VBase;
    typedef typename Halfedge::Base               HBase;
    typedef typename Face::Base                   FBase;

    // Handles and Iterators
    typedef typename HDS::Vertex_handle           Vertex_handle;
    typedef typename HDS::Halfedge_handle         Halfedge_handle;
    typedef typename HDS::Face_handle             Face_handle;
    typedef typename HDS::Vertex_iterator         Vertex_iterator;
    typedef typename HDS::Halfedge_iterator       Halfedge_iterator;
    typedef typename HDS::Face_iterator           Face_iterator;

    typedef typename HDS::Vertex_const_handle     Vertex_const_handle;
    typedef typename HDS::Halfedge_const_handle   Halfedge_const_handle;
    typedef typename HDS::Face_const_handle       Face_const_handle;
    typedef typename HDS::Vertex_const_iterator   Vertex_const_iterator;
    typedef typename HDS::Halfedge_const_iterator Halfedge_const_iterator;
    typedef typename HDS::Face_const_iterator     Face_const_iterator;

    // Auxiliary iterators for convenience
    typedef CGAL::Project_point<Vertex>           Proj_point;
    typedef CGAL::Iterator_project<Vertex_iterator, Proj_point>
                                                  Point_iterator;
    typedef CGAL::Iterator_project<Vertex_const_iterator, Proj_point,
        const Point&, const Point*>               Point_const_iterator;

    typedef CGAL::N_step_adaptor_derived<Halfedge_iterator, 2>
                                                  Edge_iterator;
    typedef CGAL::N_step_adaptor_derived<Halfedge_const_iterator, 2>
                                                  Edge_const_iterator;

    // Supported options by HDS.
    typedef typename VBase::Supports_vertex_halfedge
                                                  Supports_vertex_halfedge;
    typedef typename HBase::Supports_halfedge_prev  Supports_halfedge_prev;
    typedef typename HBase::Supports_halfedge_prev  Supports_prev;
    typedef typename HBase::Supports_halfedge_vertex
                                                  Supports_halfedge_vertex;
    typedef typename HBase::Supports_halfedge_face  Supports_halfedge_face;
    typedef typename FBase::Supports_face_halfedge  Supports_face_halfedge;

public:
    // Circulator category.
    typedef CGAL::Polyhedron_circulator_traits<Supports_prev> Ctr;
    typedef typename Ctr::iterator_category circulator_category;

    // Circulators around a vertex and around a facet.
    typedef CGAL::I_Polyhedron_facet_circ<Halfedge_handle, circulator_category>
                                         Halfedge_around_face_circulator;

    typedef CGAL::I_Polyhedron_vertex_circ<Halfedge_handle,circulator_category>
                                        Halfedge_around_vertex_circulator;

    typedef CGAL::I_Polyhedron_facet_circ<
        Halfedge_const_handle,
        circulator_category>       Halfedge_around_face_const_circulator;

    typedef CGAL::I_Polyhedron_vertex_circ<
        Halfedge_const_handle,
        circulator_category>      Halfedge_around_vertex_const_circulator;

private:
    HDS     hds;  // the boundary representation.
    Traits  m_traits;

// CREATION
public:

    PseudoT_2( const Traits& trts = Traits()) : m_traits(trts) {
        // the empty PseudoT.
	hds.faces_push_back( Face());
    }

    PseudoT_2( size_type v, size_type h, size_type f,
	       const Traits& traits = Traits())
    : hds(v,h,f), m_traits(traits) {
        // a PseudoT with storage reserved for v vertices, h
        // halfedges, and f facets. The reservation sizes are a hint for
        // optimizing storage allocation.
	hds.faces_push_back( Face());
    }

    void reserve( size_type v, size_type h, size_type f) {
        // reserve storage for v vertices, h halfedges, and f facets. The
        // reservation sizes are a hint for optimizing storage allocation.
        // If the `capacity' is already greater than the requested size
        // nothing happens. If the `capacity' changes all iterators and
        // circulators invalidates.
        hds.reserve(v,h,f);
    }

// Access Member Functions

    allocator_type get_allocator() const { return hds.get_allocator(); }

    size_type size_of_vertices() const { return hds.size_of_vertices();}
        // number of vertices.

    size_type size_of_halfedges() const { return hds.size_of_halfedges();}
        // number of all halfedges (including border halfedges).

    size_type size_of_faces() const { return hds.size_of_faces();}
        // number of faces.

    size_type capacity_of_vertices() const {
        // space reserved for vertices.
        return hds.capacity_of_vertices();
    }

    size_type capacity_of_halfedges() const {
        // space reserved for halfedges.
        return hds.capacity_of_halfedges();
    }

    size_type capacity_of_faces() const {
        // space reserved for faces.
        return hds.capacity_of_faces();
    }

    std::size_t bytes() const {
        // bytes used for the PseudoT.
        return sizeof(Self) - sizeof(HDS) + hds.bytes();
    }

    std::size_t bytes_reserved() const {
        // bytes reserved for the PseudoT.
        return sizeof(Self) - sizeof(HDS) + hds.bytes_reserved();
    }

    Vertex_iterator vertices_begin() { return hds.vertices_begin();}
        // iterator over all vertices.

    Vertex_iterator vertices_end() { return hds.vertices_end();}

    Halfedge_iterator halfedges_begin() { return hds.halfedges_begin();}
        // iterator over all halfedges

    Halfedge_iterator halfedges_end() { return hds.halfedges_end();}

    Face_iterator outer_face() { return hds.faces_begin(); }

    Face_iterator faces_begin() {
        // iterator over all faces
	Face_iterator f = hds.faces_begin();
	++f; // skip outer face
	return f;
    }

    Face_iterator faces_end() { return hds.faces_end();}

    // The constant iterators and circulators.

    Vertex_const_iterator vertices_begin() const {
        return hds.vertices_begin();
    }
    Vertex_const_iterator vertices_end() const {
        return hds.vertices_end();
    }

    Halfedge_const_iterator halfedges_begin() const {
      return hds.halfedges_begin();
    }
    Halfedge_const_iterator halfedges_end() const {
        return hds.halfedges_end();
    }

    Face_const_iterator outer_face() const { return hds.faces_begin();}
    Face_const_iterator faces_begin() const {
        // iterator over all faces
	Face_const_iterator f = hds.faces_begin();
	++f; // skip outer face
	return f;
    }
    Face_const_iterator faces_end()   const { return hds.faces_end();}

    // Auxiliary iterators for convinience
    Point_iterator       points_begin()       { return vertices_begin();}
    Point_iterator       points_end()         { return vertices_end();}

    Point_const_iterator points_begin() const { return vertices_begin();}
    Point_const_iterator points_end()   const { return vertices_end();}

    Edge_iterator        edges_begin()        { return halfedges_begin();}
        // iterator over all edges. The iterator refers to halfedges, but
        // enumerates only one of the two corresponding opposite
        // halfedges.
    Edge_iterator        edges_end()          { return halfedges_end();}
        // end of the range over all edges.

    Edge_const_iterator  edges_begin()  const { return halfedges_begin();}
    Edge_const_iterator  edges_end()    const { return halfedges_end();}

    Traits&       traits()       { return m_traits; }
    const Traits& traits() const { return m_traits; }


// Querying

    bool is_adjacent_vertex( Halfedge_const_handle v,
			     Halfedge_const_handle w) const {
	// Tests whether vertex w is adjacent to vertex v, i.e., 
	// that there exits an edge from v to w.
	Halfedge_const_handle start = v;
	do {
	    if ( v->next()->vertex() == w->vertex())
		return true;
	    v = v->next_on_vertex();
	} while( v != start);
	return false;
    }
    bool is_adjacent_face( Halfedge_const_handle f,
			   Halfedge_const_handle g) const {
	// Tests whether face g is adjacent to face f, i.e., 
	// that there exits an edge between f and g.
	Halfedge_const_handle start = f;
	do {
	    if ( f->opposite()->face() == g->face())
		return true;
	    f = f->next_on_vertex();
	} while( f != start);
	return false;
    }

// Inserting

    Vertex_handle vertices_push_back( const Vertex& v) {
        // insert single vertex
	return hds.vertices_push_back( v);
    }
    Vertex_handle vertices_push_back( const Point& p) {
        // insert single vertex
	return vertices_push_back( Vertex( p));
    }

// Erasing

    void vertices_erase( Vertex_handle v) { hds.vertices_erase( v); }

    void clear() {
        // removes all vertices, halfedges, and facets.
	hds.clear();
	hds.faces_push_back( Face());
    }

    void edges_faces_clear() {
        // removes all edges and faces
	hds.faces_clear();
	hds.edges_clear();
	hds.faces_push_back( Face());
	for ( Vertex_iterator i = hds.vertices_begin();
	      i != hds.vertices_end(); ++i) {
	    i->VBase::set_halfedge( Halfedge_handle());
	}
    }

// Euler Operators (Combinatorial Modifications)
//
// The following Euler operations modify consistently the combinatorial
// structure of the polyhedral surface. The geometry remains unchanged.

    Halfedge_handle split_facet( Halfedge_handle h, Halfedge_handle g) {
        // split the facet incident to `h' and `g' into two facets with
        // new diagonal between the two vertices denoted by `h' and `g'
        // respectively. The second (new) facet is a copy of the first
        // facet. It returns the new diagonal. The time is proportional to
        // the distance from `h' to `g' around the facet. Precondition:
        // `h' and `g' are incident to the same facet. `h != g' (no
        // loops). `h->next() != g' and `g->next() != h' (no multi-edges).
        reserve( size_of_vertices(),
                 2 + size_of_halfedges(),
                 1 + size_of_facets());
        CGAL::HalfedgeDS_decorator<HDS> D(hds);
        CGAL_precondition( D.get_face(h) == D.get_face(g));
        CGAL_precondition( h != g);
        CGAL_precondition( h != g->next());
        CGAL_precondition( h->next() != g);
        return D.split_face( h, g);
    }

    Halfedge_handle join_facet( Halfedge_handle h) {
        // join the two facets incident to h. The facet incident to
        // `h->opposite()' gets removed. Both facets might be holes.
        // Returns the predecessor of h. The invariant `join_facet(
        // split_facet( h, g))' returns h and keeps the PseudoT
        // unchanged. The time is proportional to the size of the facet
        // removed and the time to compute `h.prev()'. Precondition:
        // `HDS' supports removal of facets. The degree of both
        // vertices incident to h is at least three (no antennas).
        CGAL::HalfedgeDS_decorator<HDS> D(hds);
        CGAL_precondition( CGAL::circulator_size(h->vertex_begin())
                           >= size_type(3));
        CGAL_precondition( CGAL::circulator_size(h->opposite()->vertex_begin())
                           >= size_type(3));
        return D.join_face(h);
    }

    Halfedge_handle split_vertex( Halfedge_handle h, Halfedge_handle g) {
        // split the vertex incident to `h' and `g' into two vertices and
        // connects them with a new edge. The second (new) vertex is a
        // copy of the first vertex. It returns the new edge. The time is
        // proportional to the distance from `h' to `g' around the vertex.
        // Precondition: `h' and `g' are incident to the same vertex. `h
        // != g' (no antennas). `h->next() != g' and `g->next() != h'.
        reserve( 1 + size_of_vertices(),
                 2 + size_of_halfedges(),
                 size_of_facets());
        CGAL::HalfedgeDS_decorator<HDS> D(hds);
        CGAL_precondition( D.get_vertex(h) == D.get_vertex(g));
        CGAL_precondition( h != g);
        return D.split_vertex( h, g);
    }

    Halfedge_handle join_vertex( Halfedge_handle h) {
        // join the two vertices incident to h. The vertex denoted by
        // `h->opposite()' gets removed. Returns the predecessor of h. The
        // invariant `join_vertex( split_vertex( h, g))' returns h and
        // keeps the PseudoT unchanged. The time is proportional to
        // the degree of the vertex removed and the time to compute
        // `h.prev()'.
        // Precondition: `HDS' supports removal of vertices. The size of
        // both facets incident to h is at least four (no multi-edges)
        CGAL::HalfedgeDS_decorator<HDS> D(hds);
        CGAL_precondition( CGAL::circulator_size( h->facet_begin())
                           >= size_type(4));
        CGAL_precondition( CGAL::circulator_size( h->opposite()->facet_begin())
                           >= size_type(4));
        return D.join_vertex(h);
    }

    Halfedge_handle split_edge( Halfedge_handle h) {
        return split_vertex( h->prev(), h->opposite())->opposite();
    }

    Halfedge_handle create_center_vertex( Halfedge_handle h) {
        CGAL::HalfedgeDS_decorator<HDS> D(hds);
        CGAL_assertion( CGAL::circulator_size( h->facet_begin())
                        >= size_type(3));
        return D.create_center_vertex(h);
    }

    Halfedge_handle erase_center_vertex( Halfedge_handle h) {
        CGAL::HalfedgeDS_decorator<HDS> D(hds);
        return D.erase_center_vertex(h);
    }

// Special Operations

    void delegate( CGAL::Modifier_base<HDS>& modifier) {
        // calls the `operator()' of the `modifier'. Precondition: The
        // `modifier' returns a consistent representation.
        modifier( hds);
        CGAL_expensive_postcondition( is_valid());
    }
    bool is_valid( bool verb = false, int level = 0) const;
        // checks the combinatorial consistency.

    bool is_pseudo_triangulation( bool verb = false) const;

// Algorithms for Pseudo Triangulations
    void triangulate( bool pseudo = true, bool reverse = false);
    void greedy_triangulate();
    void bounded_vertex_triangulate();

private:
    void jarvis_march( std::list<Vertex_handle>& vertices,
		       Face_handle o_face,
		       Halfedge_handle& hull) {
	// Precond: no duplicate points, no collinear points, exact predicates
	Less_xy_2  less_xy  = traits().less_xy_2_object();
	Leftturn_2 leftturn = traits().leftturn_2_object();
	CGAL::HalfedgeDS_decorator<HDS> decorator(hds);
	// special cases of few vertices
	if ( vertices.size() < 2)
	    return;
	if ( vertices.size() == 2) {
	    Vertex_handle v1 = vertices.front();
	    Vertex_handle v2 = vertices.back();
	    hull = hds.edges_push_back( Halfedge(), Halfedge());
	    decorator.close_tip( hull->opposite(), v1);
	    decorator.close_tip( hull, v2);
	    decorator.set_face( hull->opposite(), o_face);
	    decorator.set_face( hull, o_face);
	    decorator.set_vertex_halfedge( hull->opposite());
	    decorator.set_vertex_halfedge( hull);
	    decorator.set_face_halfedge( hull);
	    return;
	}
	// general case of three or more vertices creates inner face
	Face_handle i_face = hds.faces_push_back( Face());
	// find lex-smallest vertex in curr
	typedef typename std::list<Vertex_handle>::iterator Iterator;
	Iterator curr = vertices.begin();    
	for ( Iterator i = vertices.begin(); i != vertices.end(); ++i) {
	    if ( less_xy( (*i)->point(), (*curr)->point()))
		curr = i;
	}
	std::swap( *curr, vertices.front());
	Vertex_handle vcurr = vertices.front();
	// first halfedge
	Halfedge_handle newh = hds.edges_push_back( Halfedge(), Halfedge());
	hull = newh;
	decorator.close_tip( newh->opposite(), vcurr);
	decorator.set_face( newh->opposite(), o_face);
	decorator.set_face( newh, i_face);
	decorator.set_vertex_halfedge( newh->opposite());
	decorator.set_face_halfedge( newh->opposite());
	decorator.set_face_halfedge( newh);
	// wrap
	while (1) {
	    Iterator next = vertices.end();
	    --next;
	    // search next vertex on hull
	    for ( Iterator k = vertices.begin(); k != vertices.end(); ++k) {
		if (leftturn( (*k)->point(), (*next)->point(), vcurr->point()))
		    next = k;
	    }
	    if ( next == vertices.begin()) {
		// found starting point, terminate
		decorator.insert_tip( newh, hull->opposite());
		break;
	    }
	    // found new segment from curr to next
	    decorator.close_tip( newh, *next);
	    Halfedge_handle hh = hds.edges_push_back( Halfedge(), Halfedge());
	    decorator.insert_tip( hh->opposite(), newh);
	    newh = hh;
	    decorator.set_face( newh->opposite(), o_face);
	    decorator.set_face( newh, i_face);
	    decorator.set_vertex_halfedge( newh->opposite());	
	    // continue
	    vcurr = *next;
	    vertices.erase(next);
	}
	// remove lex-smallest vertex as well
	vertices.pop_front();
    }

    void bounded_vertex_triangulate( 
	Halfedge_handle               hull,
	size_t                        hull_size,
	std::vector<Halfedge_handle>& loaded_vertices,
	std::list<Vertex_handle>&     vertices) {
	if ( hull_size < 3)
	    return;
	if ( hull_size == 3) {
	    // shelling:
	    Halfedge_handle new_hull;
	    jarvis_march( vertices, hull->face(), new_hull);
	    
	    return;
	}
    }

public:

    bool is_flippable( Halfedge_const_handle h) const {
	// tests for a valid flip in a triangulation
	Leftturn_2 leftturn = traits().leftturn_2_object();
	Halfedge_const_handle g = h->opposite();
	return h->is_triangle()
	    && h->opposite()->is_triangle()
	    && leftturn( g->next()->vertex()->point(),
			 h->vertex()->point(),
			 h->next()->vertex()->point())
	    && leftturn( h->next()->vertex()->point(),
			 g->vertex()->point(),
			 g->next()->vertex()->point());
    }    

    Halfedge_handle flip( Halfedge_handle h) {
	// flip in a triangulation. Returns h rotated ccw.
	CGAL_precondition( is_flippable( h));
	CGAL::HalfedgeDS_decorator<HDS> decorator(hds);
	return decorator.flip_edge(h);
    }
    std::pair< Halfedge_handle, Halfedge_handle>
    find_pseudo_triangle_flip_general( Halfedge_handle h) {
	// Find the vertices for the flipped diagonal.
	// Note: the halfedges denoting these vertices are not
	// necessarily incident to the correct face, but they are
	// safe to use even if the old diagonal h gets removed.
	// Returns new diagonal, rotated counterclockwise.
	CGAL_precondition( h->is_pseudo_triangle());
	CGAL_precondition( h->opposite()->is_pseudo_triangle());
	Rotate_ccw_less_2 rotate_ccw_less= traits().rotate_ccw_less_2_object();
	Leftturn_2        leftturn       = traits().leftturn_2_object();
	Halfedge_handle diag = h;
	Halfedge_handle g = h->opposite();
	bool reverse_h = h->is_corner();
	bool reverse_g = g->is_corner();
	CGAL_assertion_code( int terminate_count = INT_MAX;)
	do {
	    // decide on next event (rotational sweep), g and h start with
	    // reverse orientation to each other.
	    if ( (reverse_h != reverse_g) 
		 != rotate_ccw_less( h->vertex()->point(),
				     h->next()->vertex()->point(),
				     g->next()->vertex()->point(),
				     g->vertex()->point())) {
		// test advance of g as solution
		if ( reverse_g != leftturn( h->vertex()->point(),
					    g->vertex()->point(),
					    g->next()->vertex()->point()))
		    break;
		// not a solution yet, advance g
		g = g->next();
		if ( g->is_corner())
		    reverse_g = ! reverse_g;
	    } else {
		// test advance of h as solution
		if ( reverse_h != leftturn( g->vertex()->point(),
					  h->vertex()->point(),
					  h->next()->vertex()->point()))
		    break;
		// not a solution yet, advance h
		h = h->next();
		if ( h->is_corner())
		    reverse_h = ! reverse_h;
	    }
	    CGAL_assertion( -- terminate_count);
	} while( 1);
	// Make solution safe for the removal of the old diagonal
	if ( h == diag)
	    h = h->opposite()->prev();
	if ( g == diag->opposite())
	    g = g->opposite()->prev();
	return  std::make_pair( h, g);
    }
    std::pair< Halfedge_handle, Halfedge_handle>
    find_pseudo_triangle_flip( Halfedge_handle h) {
	// Find the vertices for the flipped diagonal.
	// Note: the halfedges denoting these vertices are not
	// necessarily incident to the correct face, but they are
	// safe to use even if the old diagonal h gets removed.
	// Returns new diagonal, rotated counterclockwise.
	CGAL_precondition( h->is_pseudo_triangle());
	CGAL_precondition( h->opposite()->is_pseudo_triangle());
	Rotate_ccw_less_2 rotate_ccw_less= traits().rotate_ccw_less_2_object();
	Leftturn_2        leftturn       = traits().leftturn_2_object();
	Halfedge_handle diag = h;
	Halfedge_handle g = h->opposite();
        // Check for special case of one or two triangles:
        if ( h->next()->next()->next() == h) {
            if ( g->next()->next()->next() == g)
                return std::make_pair( h->next(), g->next());
            // h is incident to a triangle, g to a pseudo-triangle
            bool reverse_h = true;
            bool reverse_g = g->is_corner();
            CGAL_assertion_code( int terminate_count = INT_MAX;)
            do {
                // decide on next event (rotational sweep), g and h start
                // with reverse orientation to each other.
                if ( (reverse_h != reverse_g) 
                     != rotate_ccw_less( h->vertex()->point(),
                                         h->next()->vertex()->point(),
                                         g->next()->vertex()->point(),
                                         g->vertex()->point())) {
                    // test advance of g as solution
                    if ( reverse_g != leftturn( h->vertex()->point(),
                                                g->vertex()->point(),
                                                g->next()->vertex()->point()))
                        break;
                    // not a solution yet, advance g
                    g = g->next();
                    if ( g->is_corner())
                        reverse_g = ! reverse_g;
                } else {
                    // test advance of h as solution
                    if ( reverse_h != leftturn( g->vertex()->point(),
                                                h->vertex()->point(),
                                                h->next()->vertex()->point()))
                        break;
                    // not a solution yet, advance h
                    h = h->next();
                    reverse_h = ! reverse_h;
                }
                CGAL_assertion( -- terminate_count);
            } while( 1);
            // Make solution safe for the removal of the old diagonal
            if ( h == diag)
                h = h->opposite()->prev();
            if ( g == diag->opposite())
                g = g->opposite()->prev();
            return  std::make_pair( h, g);
        }
        if ( g->next()->next()->next() == g) {
            // g is incident to a triangle, h to a pseudo-triangle
            bool reverse_h = h->is_corner();
            bool reverse_g = true;
            CGAL_assertion_code( int terminate_count = INT_MAX;)
            do {
                // decide on next event (rotational sweep), g and h start
                // with reverse orientation to each other.
                if ( (reverse_h != reverse_g) 
                     != rotate_ccw_less( h->vertex()->point(),
                                         h->next()->vertex()->point(),
                                         g->next()->vertex()->point(),
                                         g->vertex()->point())) {
                    // test advance of g as solution
                    if ( reverse_g != leftturn( h->vertex()->point(),
                                                g->vertex()->point(),
                                                g->next()->vertex()->point()))
                        break;
                    // not a solution yet, advance g
                    g = g->next();
                    reverse_g = ! reverse_g;
                } else {
                    // test advance of h as solution
                    if ( reverse_h != leftturn( g->vertex()->point(),
                                                h->vertex()->point(),
                                                h->next()->vertex()->point()))
                        break;
                    // not a solution yet, advance h
                    h = h->next();
                    if ( h->is_corner())
                        reverse_h = ! reverse_h;
                }
                CGAL_assertion( -- terminate_count);
            } while( 1);
            // Make solution safe for the removal of the old diagonal
            if ( h == diag)
                h = h->opposite()->prev();
            if ( g == diag->opposite())
                g = g->opposite()->prev();
            return  std::make_pair( h, g);
        }
        // General case of at least one true pseudo-triangle
        return find_pseudo_triangle_flip_general(h);
    }
    std::pair< Halfedge_const_handle, Halfedge_const_handle>
    find_pseudo_triangle_flip( Halfedge_const_handle h) const {
        Halfedge_handle& g = reinterpret_cast<Halfedge_handle&>(h);
        std::pair< Halfedge_handle, Halfedge_handle> result =
            (const_cast< Self*>(this))->find_pseudo_triangle_flip(g);
        return std::pair< Halfedge_const_handle, Halfedge_const_handle>(
            result.first, result.second);
    }
    std::pair< Halfedge_handle, Halfedge_handle>
    find_pseudo_quadrangle_diagonale( Halfedge_handle h, Halfedge_handle g) {
	// Find the vertices for the next quadrangle diagonal.
	// Returns the diagonal, rotated counterclockwise.
	Rotate_ccw_less_2 rotate_ccw_less =traits().rotate_ccw_less_2_object();
	Leftturn_2   leftturn   = traits().leftturn_2_object();
	bool reverse_h = h->is_corner();
	bool reverse_g = g->is_corner();
	CGAL_assertion_code( int terminate_count = INT_MAX;)
	do {
	    // decide on next event (rotational sweep), g and h start with
	    // reverse orientation to each other.
	    if ( (reverse_h != reverse_g) 
		 != rotate_ccw_less( h->vertex()->point(),
				     h->next()->vertex()->point(),
				     g->next()->vertex()->point(),
				     g->vertex()->point())) {
		// test advance of g as solution
		if ( reverse_g != leftturn( h->vertex()->point(),
					    g->vertex()->point(),
					    g->next()->vertex()->point()))
		    break;
		// not a solution yet, advance g
		g = g->next();
		if ( g->is_corner())
		    reverse_g = ! reverse_g;
	    } else {
		// test advance of h as solution
		if ( reverse_h != leftturn( g->vertex()->point(),
					  h->vertex()->point(),
					  h->next()->vertex()->point()))
		    break;
		// not a solution yet, advance h
		h = h->next();
		if ( h->is_corner())
		    reverse_h = ! reverse_h;
	    }
	    CGAL_assertion( -- terminate_count);
	} while( 1);
	return  std::make_pair( h, g);
    }
    void split_pseudo_quadrangle( Halfedge_handle h) {
        CGAL_precondition( h->size_of_corners() == 4);
        CGAL::HalfedgeDS_decorator<HDS> decorator(hds);
        Less_xy_2  less_xy  = traits().less_xy_2_object();
        while ( ! h->is_corner())
            h = h->next();
        Halfedge_handle g = h;
        do {
            g = g->next();
        } while ( ! g->is_corner());
        do {
            g = g->next();
        } while ( ! g->is_corner());
        std::pair< Halfedge_handle, Halfedge_handle> diag1 =
            find_pseudo_quadrangle_diagonale( h, g);
        // test correct slope of diag1
        if ( less_xy( diag1.second->vertex()->point(), 
                      diag1.first->vertex()->point())) {
            std::swap( h, g);
            std::swap( diag1.first, diag1.second);
        }
        do {
            h = h->next();
        } while ( ! h->is_corner());
        do {
            g = g->next();
        } while ( ! g->is_corner());
        std::pair< Halfedge_handle, Halfedge_handle> diag2 =
              find_pseudo_quadrangle_diagonale( h, g);
        CGAL_assertion( diag1 != diag2);
        if ( less_xy( diag2.second->vertex()->point(), 
                      diag2.first->vertex()->point()))
            diag1 = make_pair( diag2.second, diag2.first);
        decorator.split_face( diag1.first, diag1.second);
    }
    bool is_pseudo_flippable( Halfedge_const_handle h) const {
	// always true in a pseudo triangualtion, just make sure we 
	// have at least triangles.
	return h->is_pseudo_triangle() && h->opposite()->is_pseudo_triangle();
    }
    bool is_strict_pseudo_flippable( Halfedge_const_handle h,
				     Halfedge_const_handle& g1,
				     Halfedge_const_handle& g2) const {
	// Tests for a pseudo flip whose new diagonal does not
	// dispappear because it would be identical with an already
	// existing edge. Returns corners to new diagonal.
	g1 = h->opposite()->prev();
	g2 = h->prev();
	if ( ! is_pseudo_flippable(h))
	    return false;
	std::pair< Halfedge_const_handle, Halfedge_const_handle> diag =
	    find_pseudo_triangle_flip( h);
	g1 = diag.first;
	g2 = diag.second;
	if ( g1->vertex() == g2->vertex())
	    return false;
	if ( is_adjacent_vertex( g1, g2))
	    return false;
	return true;
    }
    bool is_strict_pseudo_flippable( Halfedge_handle h,
				     Halfedge_handle& g1,
				     Halfedge_handle& g2) {
	// Tests for a pseudo flip whose new diagonal does not
	// dispappear because it would be identical with an already
	// existing edge. Returns corners to new diagonal.
	g1 = h->opposite()->prev();
	g2 = h->prev();
	if ( ! is_pseudo_flippable(h))
	    return false;
	std::pair< Halfedge_handle, Halfedge_handle> diag =
	    find_pseudo_triangle_flip( h);
	g1 = diag.first;
	g2 = diag.second;
	if ( g1->vertex() == g2->vertex())
	    return false;
	if ( is_adjacent_vertex( g1, g2))
	    return false;
	return true;
    }
    bool is_strict_pseudo_flippable( Halfedge_const_handle h) const {
	Halfedge_const_handle dummy1;
	Halfedge_const_handle dummy2;
	return is_strict_pseudo_flippable( h, dummy1, dummy2);
    }
    Halfedge_handle pseudo_flip( Halfedge_handle h, 
				 Halfedge_handle g1,
				 Halfedge_handle g2,
				 bool reverse = false) {
	// h is the old diagonal, g1--g2 is the new diagonal.
	// rotates h counterclockwise if reverse == false, otherwise clockwise.
	CGAL_precondition( h != g1 && h->opposite() != g1);
	CGAL_precondition( h != g2 && h->opposite() != g2);
	CGAL::HalfedgeDS_items_decorator<HDS> decorator;
	Halfedge_handle h1 = h->prev();
	Halfedge_handle h2 = h->opposite()->prev();
	decorator.remove_tip( h1);
	decorator.remove_tip( h2);
	decorator.set_vertex_halfedge( h1);
	decorator.set_vertex_halfedge( h2);
	if ( reverse)
	    h = h->opposite();
	decorator.insert_tip( h, g1);
	decorator.insert_tip( h->opposite(), g2);
	decorator.set_face_in_face_loop( h, decorator.get_face( h));
	decorator.set_face_in_face_loop( h->opposite(),
					 decorator.get_face( h->opposite()));
	decorator.set_face_halfedge( h);
	decorator.set_face_halfedge( h->opposite());
	return h;
    }
    Halfedge_handle pseudo_flip( Halfedge_handle h, bool reverse = false) {
	CGAL_precondition( is_pseudo_flippable( h));
	std::pair< Halfedge_handle, Halfedge_handle> diag =
	    find_pseudo_triangle_flip( h);
	return pseudo_flip( h, diag.first, diag.second, reverse);
    }
    bool is_improving_lifting( Halfedge_const_handle h) const {
	In_circle_2  in_circle  = traits().in_circle_2_object();
	Halfedge_const_handle g1;
	Halfedge_const_handle g2;
	return is_strict_pseudo_flippable( h, g1, g2)
	    && in_circle( h->opposite()->vertex()->point(),
			  h->vertex()->point(),
			  g1->vertex()->point(),
			  g2->vertex()->point());
    }
    bool is_improving_angle( Halfedge_const_handle h, double& gamma) const {
	Angle_2  angle = traits().angle_2_object();
	Halfedge_const_handle g1;
	Halfedge_const_handle g2;
	if ( is_strict_pseudo_flippable( h, g1, g2)) {
	    const Point& p = h->vertex()->point();
	    const Point& q = h->opposite()->vertex()->point();
	    const Point& r = g1->vertex()->point();
	    const Point& s = g2->vertex()->point();
	    double alpha1 = angle( h->next()->vertex()->point(), p, q);
	    double alpha2 = angle( q, p, h->opposite()->prev()->opposite()
				   ->vertex()->point());
	    double alpha3 = angle( h->opposite()->next()->vertex()->point(),
				   q, p);
	    double alpha4 = angle( p, q, h->prev()->opposite()
				   ->vertex()->point());
	    CGAL_assertion( alpha1 > 0.0 && alpha2 > 0.0 
			    && alpha3 > 0.0 && alpha4 > 0.0);
	    double min_alpha = std::min( std::min( alpha1, alpha2),
					 std::min( alpha3, alpha4));
					 
	    double beta1 = angle( g1->next()->vertex()->point(), r, s);
	    double beta2 = angle( s, r, g1->opposite()->vertex()->point());
	    double beta3 = angle( g2->next()->vertex()->point(), s, r);
	    double beta4 = angle( r, s, g2->opposite()->vertex()->point());
	    CGAL_assertion( beta1 > 0.0 && beta2 > 0.0 
			    && beta3 > 0.0 && beta4 > 0.0);
	    double min_beta = std::min( std::min( beta1, beta2),
				        std::min( beta3, beta4));
	    gamma = min_alpha;
	    return min_beta > min_alpha;
	}
	return false; // not strict pseudo flippable
    }
    bool is_improving_angle( Halfedge_const_handle h) const {
	double dummy;
	return is_improving_angle( h, dummy);
    }
    bool pseudo_flip_shares_vertex( Halfedge_const_handle h) const {
	In_circle_2  in_circle  = traits().in_circle_2_object();
	Halfedge_const_handle g1;
	Halfedge_const_handle g2;
	return is_strict_pseudo_flippable( h, g1, g2)
	    && (    h->vertex() == g1->vertex()
		    || h->opposite()->vertex() == g1->vertex()
		    || h->vertex() == g2->vertex()
		    || h->opposite()->vertex() == g2->vertex());
    }
    Halfedge_handle vertical_ray_shoot( const Point& p) {
	// Returns the right-to-left oriented halfedge above (or on) 
	// the point p (vertical ray shooting upwards). If p is in 
	// the outer face, an arbitrary halfedge incident to the outer
	// face will be reported. If there are no edges at all (just
	// the outer face) Halfedge_handle() will be returned.
	// Post: return->face() contains p.
	Less_xy_2  less_xy  = traits().less_xy_2_object();
	Leftturn_2 leftturn   = traits().leftturn_2_object();
	Halfedge_handle h_above = outer_face()->halfedge();
	for ( Edge_iterator e = edges_begin(); e != edges_end(); ++e) {
	    Halfedge_iterator h = e;
	    if ( ! less_xy( h->vertex()->point(), 
			  h->opposite()->vertex()->point()))
		++h;
	    if ( less_xy( h->vertex()->point(), p) 
		 && less_xy( p, h->opposite()->vertex()->point())) {
		// segment of h overlaps p in x-range.
		if ( leftturn( h->opposite()->vertex()->point(),
			       h->vertex()->point(),
			       p)) {
		    // segment is above p
		    if ( h_above == outer_face()->halfedge()) {
			// h is first halfedge to be found above p
			h_above = h;
		    } else {
			// compare h with h_above, knowing they don't 
			// intersect but might touch, use left end-points
			const Point& p =h->vertex()->point();
			const Point& q =h->opposite()->vertex()->point();
			const Point& r =h_above->vertex()->point();
			const Point& s =h_above->opposite()->vertex()->point();
			if ( less_xy( p, r)) {
			    // r is within the x-range of pq.
			    if ( leftturn( p, q, r))
				h_above = h;
			} else if ( less_xy( r, p)) {
			    // p is within the x-range of rs.
			    if ( leftturn( s, r, p))
				h_above = h;
			} else {
			    // Ooops, left end-points are equal, compare slope
			    if ( leftturn( p, q, s)) 
				// make the smaller slope new halfedge
				h_above = h;
			}
		    }
		}
	    }
	}
	return h_above;
    }
    Halfedge_const_handle vertical_ray_shoot( const Point& p) const {
	return ((Self*)this)->vertical_ray_shoot(p);
    }
    size_t max_degree() const {
	size_t degree = 0;
	for ( Vertex_const_iterator i = vertices_begin(); 
	      i != vertices_end();
	      ++i) {
	    size_t d = i->degree();
	    if ( degree < d)
		degree = d;
	}
	return degree;
    }
    size_t bounded_max_degree( size_t bound) const {
	// returns max_degree() if this is smaller than bound,
	// otherwise returns bound.
	size_t degree = 0;
	for ( Vertex_const_iterator i = vertices_begin(); 
	      i != vertices_end();
	      ++i) {
	    size_t d = i->degree();
	    if ( degree < d) {
		degree = d;
		if ( degree >= bound)
		    return bound;
	    }
	}
	return degree;
    }
};


#ifndef CGAL_CFG_NO_TMPL_IN_TMPL_PARAM
#define PTW__PSEUDOT_TMPL_DECL template < class TR, class Itms, \
                              template < class T, class I, class A> class HD, \
                              class Alloc>
#else
#define PTW__PSEUDOT_TMPL_DECL template < class TR, class Itms, class HD, \
                              class Alloc>
#endif


PTW__PSEUDOT_TMPL_DECL
bool
PseudoT_2<TR,Itms,HD,Alloc>:: is_valid( bool verb, int level) const {
    CGAL::Verbose_ostream verr(verb);
    verr << "begin PTW::PseudoT_2<...>::is_valid( verb=true, "
                      "level = " << level << "):" << std::endl;
    CGAL::HalfedgeDS_const_decorator<HDS> D(hds);
    bool valid = D.is_valid( verb, 3);
    if ( size_of_vertices() > 2) {
	// All halfedges.
	Halfedge_const_iterator i   = halfedges_begin();
	Halfedge_const_iterator end = halfedges_end();
	size_type  n = 0;
	for( ; valid && (i != end); ++i) {
	    verr << "halfedge " << n << std::endl;
	    // At least triangular facets and distinct geometry.
	    valid = valid && ( i->next() != i);
	    valid = valid && ( i->next()->next() != i);
	    valid = valid && ( ! check_tag( Supports_halfedge_vertex()) ||
			       D.get_vertex(i) != D.get_vertex(i->opposite()));
	    valid = valid && ( ! check_tag( Supports_halfedge_vertex()) ||
			       D.get_vertex(i) != D.get_vertex(i->next()));
	    valid = valid && ( ! check_tag( Supports_halfedge_vertex()) ||
			D.get_vertex(i) != D.get_vertex(i->next()->next()));
	    if ( ! valid) {
		verr << "    incident facet is not at least a triangle."
		     << std::endl;
		break;
	    }
	    // Distinct facets on each side of an halfegde.
	    valid = valid && ( ! check_tag( Supports_halfedge_face()) ||
			       D.get_face(i) != D.get_face(i->opposite()));
	    if ( ! valid) {
		verr << "    both incident facets are equal." << std::endl;
		break;
	    }
	    ++n;
	}
	valid = valid && (n == size_of_halfedges());
	if ( n != size_of_halfedges())
	    verr << "counting halfedges failed." << std::endl;
    }
    if ( valid && level > 0 && ! is_pseudo_triangulation( verb))
	valid = false;
    verr << "end of PTW::PseudoT_2<...>::is_valid(): structure is "
         << ( valid ? "valid." : "NOT VALID.") << std::endl;
    return valid;
}

PTW__PSEUDOT_TMPL_DECL
bool
PseudoT_2<TR,Itms,HD,Alloc>:: is_pseudo_triangulation( bool verb) const {
    CGAL::HalfedgeDS_const_decorator<HDS> D(hds);
    CGAL::Verbose_ostream verr(verb);
    verr << "begin PTW::PseudoT_2<...>::is_pseudo_triangulation( verb=true):"
	 << std::endl;
    bool valid = true;
    if ( check_tag( Supports_halfedge_face())) {
        for (Halfedge_const_iterator h = halfedges_begin(); 
             valid && h != halfedges_end(); ++h) {
            if ( D.get_face(h) != outer_face() && ! h->is_pseudo_triangle()) {
                valid = false;
                verr << "    face " << std::distance( faces_begin(), 
                                                      D.get_face(h))
                     << " is not a pseudo triangle." << endl;
            }
        }
    }
    for (Vertex_const_iterator v = vertices_begin(); 
  	 valid && v != vertices_end();
  	 ++v) {
  	if ( ! v->is_reflex()) {
  	    valid = false;
  	    verr << "    vertex " << std::distance( vertices_begin(), v)
  		 << " is not a reflex vertex." << endl;
  	}
    }
    verr << "end of PTW::PseudoT_2<...>::is_pseudo_triangulation(): "
	    "structure is " << ( valid ? "valid." : "NOT VALID.") << std::endl;
    return valid;
}

template <class Vertex_handle, class Comp>
struct Comp_vertex_handle {
    Comp comp;
    bool reverse;
    Comp_vertex_handle( const Comp& cmp, bool rev) : comp(cmp), reverse(rev) {}
    bool operator()( Vertex_handle v, Vertex_handle w) {
	if ( reverse)
	    return comp( w->point(), v->point());
	return comp( v->point(), w->point());
    }
};

PTW__PSEUDOT_TMPL_DECL
void
PseudoT_2<TR,Itms,HD,Alloc>:: triangulate( bool pseudo, bool reverse) {
    if ( hds.size_of_vertices() < 2)
	return;
    CGAL::HalfedgeDS_decorator<HDS> decorator(hds);
    edges_faces_clear();
    Less_xy_2  less_xy  = traits().less_xy_2_object();
    Leftturn_2 leftturn = traits().leftturn_2_object();
    std::vector< Vertex_handle> vertices;
    vertices.reserve( hds.size_of_vertices());
    for ( Vertex_iterator i=hds.vertices_begin(); i != hds.vertices_end(); ++i)
	vertices.push_back( i);
    Comp_vertex_handle<Vertex_handle, Less_xy_2> cmp( less_xy, reverse);
    std::sort( vertices.begin(), vertices.end(), cmp);
    typedef typename std::vector< Vertex_handle>::iterator Iterator;
    Iterator i = vertices.begin();
    Iterator j = i;
    ++i;
    // link first two nodes
    Halfedge_handle h = hds.edges_push_back( Halfedge(), Halfedge());
    decorator.close_tip( h, *i);
    decorator.close_tip( h->opposite(), *j);
    decorator.set_face( h, outer_face());
    decorator.set_face_halfedge( outer_face(), h);
    decorator.set_face( h->opposite(), outer_face());
    ++j;
    ++i;
    for ( ; i != vertices.end(); (++i, ++j)) {
	// Invariant: i is the new vertex to add.
	// j is the last vertex added. It is on the conves hull and h =
	// (*j)->halfedge() is a halfedge on the convex hull facing the
	// outer face (nothing said about its orientation with 
	// respect to i).

	// start from edge (i,j) and find right tangent (looking from i)
	// (clockwise orientation on the convex hull for the outer face!)
	Halfedge_handle g = h;
	while ( leftturn( g->opposite()->vertex()->point(), 
			  g->vertex()->point(), (*i)->point()))
	    g = g->prev();
	// Upon loop termination, g points to the right tangent end point.
	// connect right tangent
	Halfedge_handle newh = hds.edges_push_back( Halfedge(), Halfedge());
	decorator.close_tip( newh, *i);
	decorator.insert_tip( newh->opposite(), g);
	decorator.set_face( newh, outer_face());
	decorator.set_face_halfedge( outer_face(), newh);
	decorator.set_face( newh->opposite(), outer_face());
	// iterate from right tangent to the left tangent, creating edges
	// for triangulations, or just searching the left tangent
	// for pseudotriangulations
	g = newh->next();
	while ( leftturn( (*i)->point(), g->vertex()->point(), 
			  g->next()->vertex()->point())) {
	    g = g->next();
	    if ( ! pseudo) { // triangulate fan from *i
		decorator.split_face( newh, g);
		g = newh->next();
	    }
	}
	// Upon loop termination, g points to the right tangent end point.
	// Create tangent for pseudo triangulations. For triangulations,
	// the tangent has already been created in the loop above.
	if ( pseudo)
	    decorator.split_face( newh, g);
	// Done, next vertex
	h = newh;
    }
    CGAL_postcondition( is_valid());
    CGAL_postcondition( is_pseudo_triangulation());
}

PTW__PSEUDOT_TMPL_DECL
void
PseudoT_2<TR,Itms,HD,Alloc>:: greedy_triangulate() {
    if ( hds.size_of_vertices() < 2)
	return;
    CGAL::HalfedgeDS_decorator<HDS> decorator(hds);
    edges_faces_clear();
    Less_xy_2  less_xy  = traits().less_xy_2_object();
    Leftturn_2 leftturn = traits().leftturn_2_object();
    std::vector< Vertex_handle> vertices;
    vertices.reserve( hds.size_of_vertices());
    for ( Vertex_iterator i=hds.vertices_begin(); i != hds.vertices_end(); ++i)
	vertices.push_back( i);
    Comp_vertex_handle<Vertex_handle, Less_xy_2> cmp( less_xy, false);
    std::sort( vertices.begin(), vertices.end(), cmp);
    typedef typename std::vector< Vertex_handle>::iterator Iterator;
    Iterator i = vertices.end();
    --i;
    Iterator j = i;
    --i;
    // link first two nodes
    Halfedge_handle h = hds.edges_push_back( Halfedge(), Halfedge());
    h = h->opposite();
    decorator.close_tip( h, *i);
    decorator.close_tip( h->opposite(), *j);
    decorator.set_face( h, outer_face());
    decorator.set_face_halfedge( outer_face(), h);
    decorator.set_face( h->opposite(), outer_face());
    if ( hds.size_of_vertices() < 3)
	return;
    while( i != vertices.begin()) {
	--j;
	--i;
	// Invariant: i is the next vertex to the left, j is the last 
	// vertex added. h = (*j)->halfedge().

	// start from edge (i,j) and find right tangent (looking from i)
	// (clockwise orientation on the convex hull for the outer face!)
	Halfedge_handle g = h;
	while ( leftturn( g->opposite()->vertex()->point(), 
			  g->vertex()->point(), (*i)->point()))
	    g = g->prev();
	// Upon loop termination, g points to the right tangent end point.
	// connect right tangent
	Halfedge_handle newh = hds.edges_push_back( Halfedge(), Halfedge());
	newh = newh->opposite();
	decorator.close_tip( newh, *i);
	decorator.insert_tip( newh->opposite(), g);
	decorator.set_face( newh, outer_face());
	decorator.set_face_halfedge( outer_face(), newh);
	decorator.set_face( newh->opposite(), outer_face());
	h = newh;
    }
    // reverse direction
    std::swap( i, j);
    Iterator last = vertices.end();
    -- last;
    while( i != vertices.end()) {
	// Invariant: i is the next vertex to the right, j is the last 
	// vertex added. h = (*j)->halfedge() and on the outer face.

	// start from edge (i,j) and find right tangent (looking from i)
	// (clockwise orientation on the convex hull for the outer face!)
	Halfedge_handle g = h;
	while ( leftturn( g->opposite()->vertex()->point(), 
			  g->vertex()->point(), (*i)->point()))
	    g = g->prev();
	// Upon loop termination, g points to the right tangent end point.
	// connect right tangent if not already existing
	Halfedge_handle old;
	if ( last == i) // special case at the end
	    old = (*i)->halfedge();
	else
	    old = (*i)->halfedge()->opposite()->prev();
	if ( ! is_adjacent_vertex( g, old)) {
	    decorator.split_face( g, old);
	    h = g->next();
            // Integrate split of quadrangles here
            g = h->opposite();
            if ( g->size_of_corners() == 4)
                split_pseudo_quadrangle( g);
	} else {
	    // advance h
	    h = h->next();
	    CGAL_assertion( h->vertex() == *i);
	}
	++j;
	++i;
    }
    CGAL_assertion( is_valid());
    for ( Edge_iterator e = edges_begin(); e != edges_end(); ++e) {
	CGAL_warning( less_xy( e->opposite()->vertex()->point(),
				 e->vertex()->point()));
    }
    CGAL_postcondition( is_valid());
}


PTW__PSEUDOT_TMPL_DECL
void
PseudoT_2<TR,Itms,HD,Alloc>:: bounded_vertex_triangulate() {
    if ( hds.size_of_vertices() < 2)
	return;
    edges_faces_clear();

    Halfedge_handle              hull;      // handle to convex hull halfedge
    size_t                       hull_size; // size of convex hull
    std::vector<Halfedge_handle> loaded_vertices; // max 6 of loaded vertices 
    std::list<Vertex_handle>     vertices;        // inner vertices

    loaded_vertices.reserve( 6);
    for ( Vertex_iterator i=hds.vertices_begin(); i != hds.vertices_end(); ++i)
	vertices.push_back( i);
    jarvis_march( vertices, outer_face(), hull);
    hull_size = CGAL::circulator_size( hull->facet_begin());
    cerr << "hull = " << hull_size << endl << "inner = " << vertices.size() << endl;
    bounded_vertex_triangulate( hull, hull_size, loaded_vertices, vertices);
    //CGAL_postcondition( is_valid());
}

#undef PTW__PSEUDOT_TMPL_DECL

} // namespace PTW
#endif // CGAL_PSEUDOT_2_H //
// EOF //




