// ============================================================================
//
// Copyright (c) 2001  Lutz Kettner, see the Copyright file for details
//
// ----------------------------------------------------------------------------
//
// file          : PseudoT_default_traits_2.h
// revision      : $Revision: 1.2 $
// revision_date : $Date: 2001/09/03 06:13:35 $
// author(s)     : Lutz Kettner
//
// Default Traits for Pseudo Triangulations in CGAL.
// ============================================================================

#ifndef PTW_PSEUDOT_DEFAULT_TRAITS_3_H
#define PTW_PSEUDOT_DEFAULT_TRAITS_3_H 1

#include <CGAL/basic.h>
#include <CGAL/Cartesian.h>
#include <CGAL/Point_2.h>
#include <CGAL/predicates_on_points_2.h>
#include <CGAL/distance_predicates_2.h>
#include <CGAL/predicate_objects_on_points_2.h>

namespace PTW {


// Useful/missing predicate in CGAL?

// Returns true if the rotate_ccw of the oriented segment pq is less than 
// segment rs, i.e., if the rotation from pq to rs is shorter in ccw 
// direction.
template <class Rep_>
struct Rotate_ccw_less_2 {
    typedef Rep_                   R;
    typedef typename R::Point_2    Pt;
    typedef typename R::Leftturn_2 Leftturn;
    bool operator()( const Pt& p, const Pt& q, const Pt& r, const Pt& s) const{
	Pt t = p + (s - r);
	Leftturn leftturn;
	return leftturn( p, q, t);
    }
};

template <class Rep_>
struct In_circle_2 {
    typedef Rep_                   R;
    typedef typename R::Point_2    Pt;
    typedef typename R::Side_of_oriented_circle_2 Pred;
    bool operator()( const Pt& p, const Pt& q, const Pt& r, const Pt& s) const{
	Pred pred;
	return CGAL::ON_POSITIVE_SIDE == pred( p, q, r, s);
    }
};

template <class Rep_>
struct Angle_2 {
    typedef Rep_                   R;
    typedef typename R::Point_2    Pt;
    double operator()( const Pt& p, const Pt& q, const Pt& r) const{
	// angle from 0 to 2 Pi between p-q and r-q, positive in
	// ccw orientation
	typedef CGAL::Vector_2< CGAL::Cartesian< double> > Vec;
	Vec v( CGAL::to_double( (p - q).x()), CGAL::to_double( (p - q).y()));
	Vec w( CGAL::to_double( (r - q).x()), CGAL::to_double( (r - q).y()));
	// normalization not needed for atan2
	Vec ortho_v( - v.y(), v.x());
	double x = w * v;
	double y = w * ortho_v;
	double angle = atan2( y, x);
	if ( angle < 0.0)
	    angle += 2 * M_PI;
	return angle;
    }
};

template < class Rep >
class PseudoT_default_traits_2 {
public:
    // workaround for warning in egcs-2.91.66
    typedef PseudoT_default_traits_2<Rep> Self;
    Self& operator= ( const Self& ) { return *this; }

    typedef Rep                            R;
    typedef typename R::Point_2            Point_2;
    typedef Point_2                        Point;
    typedef typename R::Construct_point_2  Construct_point_2;
    typedef typename R::Less_xy_2          Less_xy_2;
    typedef typename R::Leftturn_2         Leftturn_2;
    typedef In_circle_2<R>                 In_circle_2;
    typedef Rotate_ccw_less_2<R>           Rotate_ccw_less_2;
    typedef Angle_2<R>                     Angle_2;
    typedef typename R::Less_signed_distance_to_line_2
                                           Less_signed_distance_to_line_2;

    Construct_point_2
    construct_point_2_object()          const { return Construct_point_2(); } 

    Less_xy_2     less_xy_2_object()    const { return Less_xy_2(); } 
    Leftturn_2    leftturn_2_object()   const { return Leftturn_2(); }
    In_circle_2   in_circle_2_object()  const { return In_circle_2(); }
    Rotate_ccw_less_2  rotate_ccw_less_2_object() const { 
	return Rotate_ccw_less_2(); 
    }
    Angle_2       angle_2_object()      const { return Angle_2(); }

    Less_signed_distance_to_line_2
    less_signed_distance_to_line_2_object( const Point& p, 
					   const Point& q) const {
	return Less_signed_distance_to_line_2( p, q); 
    }
};

} // namespace PTW
#endif // PTW_PSEUDOT_DEFAULT_TRAITS_3_H //
// EOF //

