// ============================================================================
//
// Copyright (c) 2001  Lutz Kettner, see the Copyright file for details
//
// ----------------------------------------------------------------------------
//
// file          : PseudoT_enumeration_2.h
// revision      : $Revision: 1.2 $
// revision_date : $Date: 2001/09/03 06:13:35 $
// author(s)     : Lutz Kettner
//
// Pseudo Triangulation Enumeration
// ============================================================================

#ifndef PTW_PSEUDOT_ENUMERATION_2_H
#define PTW_PSEUDOT_ENUMERATION_2_H 1

#include <CGAL/basic.h>
#include <PTW/PseudoT_2.h>
#include <cstddef>
#include <vector>

namespace PTW {


// Predicates for branch & bound

struct PseudoT_abort_never {
    template <class Pseudo, class Halfedge_handle>
    bool operator()( const Pseudo&, Halfedge_handle) const { return false; }
};

class PseudoT_abort_above_degree_n {
    std::size_t n;
public:
    PseudoT_abort_above_degree_n( std::size_t k = 6) : n(k) {}
    template <class Pseudo, class Halfedge_handle>
    bool operator()( const Pseudo&, Halfedge_handle h) const { 
	return h->vertex()->marked_degree() > n
	    || h->opposite()->vertex()->marked_degree() > n;
    }
};


template < class p_PseudoT, class p_Abort_branch = PseudoT_abort_never> 
class PseudoT_enumeration_2 {
public:
    typedef p_PseudoT                             PseudoT;
    typedef typename PseudoT::Traits              Traits;
    typedef p_Abort_branch                        Abort_branch;

    // Container stuff.
    typedef typename PseudoT::size_type           size_type;
    typedef typename PseudoT::difference_type     difference_type;
    typedef typename PseudoT::iterator_category   iterator_category;
    typedef typename PseudoT::Supports_removal    Supports_removal;

    // Geometry
    typedef typename Traits::Point_2              Point_2;
    typedef typename Traits::Point_2              Point;

    typedef typename Traits::Construct_point_2    Construct_point_2;
    typedef typename Traits::Less_xy_2            Less_xy_2;
    typedef typename Traits::Leftturn_2           Leftturn_2;
    typedef typename Traits::In_circle_2          In_circle_2;
    typedef typename Traits::Rotate_ccw_less_2    Rotate_ccw_less_2;
    typedef typename Traits::Less_signed_distance_to_line_2
                                                Less_signed_distance_to_line_2;

    // Items
    typedef typename PseudoT::Vertex                  Vertex;
    typedef typename PseudoT::Halfedge                Halfedge;
    typedef typename PseudoT::Face                    Face;

    // Handles and Iterators
    typedef typename PseudoT::Vertex_handle           Vertex_handle;
    typedef typename PseudoT::Halfedge_handle         Halfedge_handle;
    typedef typename PseudoT::Face_handle             Face_handle;
    typedef typename PseudoT::Vertex_iterator         Vertex_iterator;
    typedef typename PseudoT::Halfedge_iterator       Halfedge_iterator;
    typedef typename PseudoT::Edge_iterator           Edge_iterator;
    typedef typename PseudoT::Face_iterator           Face_iterator;

    typedef typename PseudoT::Vertex_const_handle     Vertex_const_handle;
    typedef typename PseudoT::Halfedge_const_handle   Halfedge_const_handle;
    typedef typename PseudoT::Face_const_handle       Face_const_handle;
    typedef typename PseudoT::Vertex_const_iterator   Vertex_const_iterator;
    typedef typename PseudoT::Halfedge_const_iterator Halfedge_const_iterator;
    typedef typename PseudoT::Edge_const_iterator     Edge_const_iterator;
    typedef typename PseudoT::Face_const_iterator     Face_const_iterator;

    // Handle halfedges sorted by slope in map
    //typedef std::multimap< double, Halfedge_handle>   Edge_map;
    //typedef typename Edge_map::iterator               Map_iterator;
    typedef std::vector<Halfedge_handle>              Stack;
    typedef typename Stack::iterator                  Iterator;
    struct Angle_cmp {
	typedef Halfedge_handle  first_argument_type;
	typedef Halfedge_handle  second_argument_type;
	typedef bool             result_type;
	bool operator()( Halfedge_handle h, Halfedge_handle g) const {
	    return h->angle() < g->angle();
	}
    };

private:
    PseudoT&      m_pseudo;
    Stack         m_stack;
    Stack         m_sort_stack;
    Abort_branch  m_abort_branch;

public:
    PseudoT_enumeration_2( PseudoT& ps) 
	: m_pseudo(ps) {}
    PseudoT_enumeration_2( PseudoT& ps, const Abort_branch& pred) 
	: m_pseudo(ps), m_abort_branch( pred) {}

    const Abort_branch& abort_branch() const { return m_abort_branch; }
    Abort_branch&       abort_branch()       { return m_abort_branch; }

    void init() { 
	m_pseudo.greedy_triangulate();
	//m_edge_map = Edge_map();
	m_stack = Stack();
	m_stack.reserve( 3 * m_pseudo.size_of_halfedges() 
			 * m_pseudo.size_of_halfedges());
	m_sort_stack = Stack();
	m_sort_stack.reserve( m_pseudo.size_of_halfedges() / 2);
	for ( Edge_iterator e = m_pseudo.edges_begin();
	      e != m_pseudo.edges_end(); ++e) {
	    e->set_marked( false);
	    if ( m_pseudo.is_strict_pseudo_flippable( e))
		m_sort_stack.push_back( e);
	}
	std::sort( m_sort_stack.begin(), m_sort_stack.end(), 
		   not2( Angle_cmp()));
	// integrate branch&bound test into left traversal,
	// i.e., test for abort whenever a new edge gets marked.
	m_stack.reserve( m_stack.size() + m_sort_stack.size());
	bool abort_branch_tag = false;
	while ( ! abort_branch_tag && ! m_sort_stack.empty()) {
	    m_sort_stack.back()->set_marked( true);
	    m_stack.push_back( m_sort_stack.back());
	    m_sort_stack.pop_back();
	    if ( m_abort_branch( m_pseudo, m_stack.back()))
		abort_branch_tag = true;
	}
	if ( abort_branch_tag)
	    next(); // aborted, continue popping the last edge
        // a pseudo triangulation has been found, report it
    }
    bool is_done() const {
	return m_stack.empty();
    }
    void next() {
	// reworked to 36% speedup, storing diagonal on cache and avoid
	// flipping when immediately clear that it is above the angle.
	while ( ! is_done()) {
	    Halfedge_handle h = m_stack.back();
	    m_stack.pop_back();
	    if ( h->is_marked()) {
		h->set_marked( false);
		std::pair< Halfedge_handle, Halfedge_handle> diag =
		    m_pseudo.find_pseudo_triangle_flip( h);
		if ( diag.first->vertex()->point().x() 
		     >= diag.second->vertex()->point().x()) {
		    // flip only if angle <= M_PI, keep old diagonal on stack
		    m_stack.push_back( h->prev());             // second
		    m_stack.push_back( h->opposite()->prev()); // first
		    m_pseudo.pseudo_flip( h, diag.first, diag.second);
		    m_stack.push_back( h);
		    m_sort_stack.push_back(h);
		    std::sort( m_sort_stack.begin(), m_sort_stack.end(), 
			       not2( Angle_cmp()));
		    // integrate branch&bound test into left traversal,
		    // i.e., test for abort whenever a new edge gets marked.
		    m_stack.reserve( m_stack.size() + m_sort_stack.size());
		    bool abort_branch_tag = false;
		    while ( ! abort_branch_tag && ! m_sort_stack.empty()) {
			m_sort_stack.back()->set_marked( true);
			m_stack.push_back( m_sort_stack.back());
			m_sort_stack.pop_back();
			if ( m_abort_branch( m_pseudo, m_stack.back()))
			    abort_branch_tag = true;
		    }
		    if ( abort_branch_tag)
			continue; // aborted, continue popping the last edge
		    break; // a pseudo triangulation has been found, report it
		} else {
		    // diagonal flips beyond angle of pi, backtrack
		    m_sort_stack.push_back( h);
		}
	    } else {
		// popped unmarked edge, i.e., we need to flip back h using 
		// the endpoints of the previous diagonal stored on the stack
		Halfedge_handle first = m_stack.back();
		m_stack.pop_back();
		Halfedge_handle second = m_stack.back();
		m_stack.pop_back();
		m_pseudo.pseudo_flip( h, first, second);
	    }
	}
    }
};


} // namespace PTW
#endif // PTW_PSEUDOT_ENUMERATION_2_H //
// EOF //

