// ============================================================================
//
// Copyright (c) 2001  Lutz Kettner, see the Copyright file for details
//
// ----------------------------------------------------------------------------
//
// file          : count_pseudo_triang.C
// revision      : $Revision: 1.2 $
// revision_date : $Date: 2001/09/03 06:36:55 $
// author(s)     : Lutz Kettner
//
// Reads a point configuration, or a whole file of configurations
// and counts the number of minimal pseudo triangulations for it.
// ============================================================================

// Oswin Aichholzer's data base works with 16 bit unsigned integers, thus
// 53 bit double arithmetic ensures exact answers for all predicates
// needed in counting pseudo triangulations.

#include <CGAL/Cartesian.h>
#include <CGAL/HalfedgeDS_vector.h>
#include <CGAL/IO/Verbose_ostream.h>
#include <CGAL/Timer.h>
#include <CGAL/Real_timer.h>

#include <PTW/Point_order_types_2.h>
#include <PTW/PseudoT_default_traits_2.h>
#include <PTW/PseudoT_2.h>
#include <PTW/PseudoT_enumeration_2.h>

#include <cstring>
#include <limits.h>


using namespace std;

typedef double                                      NT;
typedef CGAL::Cartesian< NT>                        Kernel;
typedef Kernel::Point_2                             Point;

typedef PTW::PseudoT_default_traits_2<Kernel>       Traits;

// No faces, no prev pointer
template <class Refs, class Traits_>
class My_halfedge : public CGAL::HalfedgeDS_halfedge_base<Refs, 
        CGAL::Tag_false, CGAL::Tag_true, CGAL::Tag_false> {
public:
    typedef Traits_                          Traits;
    My_halfedge() {}
};

template <class Refs, class Traits_>
class My_face : public CGAL::HalfedgeDS_face_base<Refs, CGAL::Tag_false> {
public:
    typedef Traits_                     Traits;
    My_face() {}
};

class My_items {
public:
    template < class Refs, class Traits>
    struct Vertex_wrapper {
        typedef PTW::PseudoT_vertex_2< Refs, Traits>    Vertex;
    };
    template < class Refs, class Traits>
    struct Halfedge_wrapper {
        typedef My_halfedge< Refs, Traits>  Halfedge;
    };
    template < class Refs, class Traits>
    struct Face_wrapper {
        typedef My_face< Refs, Traits>      Face;
    };
};

//typedef PTW::PseudoT_2< Traits, 
//                        PTW::PseudoT_items_2,
//                        CGAL::HalfedgeDS_default>   PseudoT;
typedef PTW::PseudoT_2< Traits, 
//                        PTW::PseudoT_items_2,
                        My_items,
                        CGAL::HalfedgeDS_vector>    PseudoT;


typedef PTW::PseudoT_enumeration_2<PseudoT>         Enumerator;

typedef PseudoT::Point_iterator                     Point_iterator;
typedef PTW::Point_order_types_2< Point_iterator, Kernel>  
                                                    Point_order_types_2;


const int MAX_POINTS = 255; // required to reserve space in vector based HDS

// global variables
// ----------------------------------------------------------------
bool verbose = false;
long start_n = 0;
long range_n = LONG_MAX;

double           user_sec = 0.0; // avoid wrap-around of time with big counters
double           real_sec = 0.0; // measuring time in seconds
CGAL::Timer      t_user;
CGAL::Real_timer t_real;

PseudoT          pseudo( MAX_POINTS, 4*MAX_POINTS-6, MAX_POINTS-1);
Enumerator       enumerator( pseudo);


// feedback: small helper function for visual feedback and adjusting timer
// -----------------------------------------------------------------------
void feedback( long counter) {
    if ( t_user.time() > 500.0) {
	user_sec += t_user.time();
	t_user.reset(); // avoids wrap-around of timer
    }
    if ( t_real.time() > 500.0) {
	real_sec += t_real.time();
	t_real.reset(); // avoids wrap-around of timer
    }
    long user_s = long( user_sec + t_user.time());
    long real_s = long( real_sec + t_real.time());
    CGAL::Verbose_ostream verr( verbose);
    verr << "\rRunning ... " << counter << " counter, "
	 << user_s/60 << 'm' << user_s%60 << "s user, "
	 << real_s/60 << 'm' << real_s%60 << "s real time."
	 << flush;
}

// count_single: read Ascii double file format with leading counter
// ----------------------------------------------------------------
long count_pseudo_triang() {
    // use enumerator to count pseudo triangulations of the points in pseudo
    long counter = 0;
    enumerator.init(); // clears also edges and faces in pseudo
    while ( ! enumerator.is_done()) {
	enumerator.next();
	++ counter;
	if ( (counter % 50000) == 0)
	    feedback( counter);
    }
    return counter;
}

// count_single: read Ascii double file format with leading counter
// ----------------------------------------------------------------
int count_single( const char* filename) {
    double x, y;
    ifstream in( filename);
    while ( in >> x >> y) {
	pseudo.vertices_push_back( Point( x, y));
    }
    if ( ! in.eof()) {
	cerr << "Error: reading points from `" << filename << "' failed."
	     << endl;
	return 1;
    }
    // use enumerator to count pseudo triangulations of the points in pseudo
    long counter = 0;
    enumerator.init(); // clears also edges and faces in pseudo
    while ( ! enumerator.is_done()) {
	enumerator.next();
	++ counter;
	if ( (counter % 50000) == 0)
	    feedback( counter);
    }
    cerr << '\r' << counter 
	 << " pseudo triangulations.                                          "
	 << endl;
    return 0;
}


// count_all: read binary point order type file and count all
// ----------------------------------------------------------------
int count_all( const char* filename, int n) {
    CGAL_precondition( n > 2);
    for ( int i = 0; i != n; ++i)
	pseudo.vertices_push_back( Point( i, i*i)); // placeholder points
    Point_order_types_2 order_types( filename, n, pseudo.points_begin());
    order_types.seek( start_n);
    int  max_triang = 0;
    int  min_triang = INT_MAX;
    long counter    = 0;
    while( counter < range_n && order_types.read()) {
	++counter;
	if (( counter % 50) == 0)
	    feedback( counter);
	// use enumerator to count pseudo triang. of the points in pseudo
	int triang = 0;
	enumerator.init(); // clears also edges and faces in pseudo
	while ( ! enumerator.is_done()) {
	    enumerator.next();
	    ++ triang;
	}
	if ( triang > max_triang)
	    max_triang = triang;
	if ( triang < min_triang)
	    min_triang = triang;
	// write entry per point order type
	switch ( n) {
	case 3:
	case 4:
	case 5:
	case 6:
	    {
		CGAL::UInteger8 i = CGAL::UInteger8(triang);
		cout << i;
		break;
	    }
	case 7:
	case 8:
	case 9:
	    {
		CGAL::UInteger16 i = CGAL::UInteger16(triang);
		// swap byte order on big endian machines
		#ifdef CGAL_BIG_ENDIAN 
		char* p = reinterpret_cast<char*>( &i);
		std::swap( p[0], p[1]);
		#endif
		cout.write(& i, 2);
		break;
	    }
	default:
	    {
		CGAL::UInteger32 i = CGAL::UInteger32(triang);
		// swap byte order on big endian machines
		#ifdef CGAL_BIG_ENDIAN 
		char* p = reinterpret_cast<char*>( &i);
		std::swap( p[0], p[3]);
		std::swap( p[1], p[2]);
		#endif
		cout.write(& i, 4);
		break;
	    }
	}
    }
    cerr << '\r' << counter 
	 << " point order type configurations.                                "
	 << endl;
    cerr << min_triang << " # min pseudo triang., "
	 << max_triang << " # max pseudo triang." << endl;
    return 0;
}


// main function with standard unix commandline arguments
// ----------------------------------------------------------------
int main( int argc, char **argv) {
    const int min_parameters = 1;
    const int max_parameters = 2;
    int n = 0; // number of parameters
    char *parameters[ max_parameters];
    bool help = false;
    bool all  = false;
    for (int i = 1; i < argc; i++) { // check commandline options
	if ( strcmp( "-all", argv[i]) == 0)
            all = true;
	else if ( strcmp( "-start", argv[i]) == 0) {
            ++i;
            if ( i < argc)
                start_n = atol( argv[i]);
            else {
                cerr << "Error: option -start needs argument." << endl;
                n = max_parameters + 1;
            }
	} else if ( strcmp( "-range", argv[i]) == 0) {
            ++i;
            if ( i < argc)
                range_n = atol( argv[i]);
            else {
                cerr << "Error: option -range needs argument." << endl;
                n = max_parameters + 1;
            }
        } else if ( strcmp( "-v", argv[i]) == 0)
            verbose = true;
        else if ( (strcmp( "-h", argv[i]) == 0) || 
	     (strcmp( "-help", argv[i]) == 0))
            help = true;
        else if ( n < max_parameters ) {
            parameters[ n++] = argv[i];
        } else {
	    ++n;
            break;
	}
    }
    int n_points = 0;
    if ( all && n > min_parameters) {
	n_points = atoi( parameters[1]);
	if ( n_points < 3 || n > 10) {
	    cerr << "Error: -all requires an integer in [3,10] as second "
		    "argument." << endl;
	    help = true;
	}
    }
    if (( (all ? n-1 : n) < min_parameters) || (n > max_parameters) || help) {
        if ( ! help)
            cerr << "Error: in parameter list" << endl;
        cerr << "Usage: count_pseudo_triang [<options>] <infile>\n\n";

	cerr << "    Counts number of pseudo triangulations for the points "
                        "in <infile> in\n";
	cerr << "    Cartesian double coordinates stored as Ascii.\n\n";
	cerr << "Usage: count_pseudo_triang -all [<options>] <infile> <n>\n\n";
	cerr << "    Counts number of pseudo triangulations for all point "
                         "configurations\n";
	cerr << "    of <n> points found in <infile> following the binary "
                         "format from\n";
	cerr << "    Oswin Aichholzer's data base of all order types of small "
                         "point sets.\n";
	cerr << "    Writes the result in the same binary format as used for "
                         "the results in\n";
	cerr << "    the data base. Results are written to stdout.\n\n";
        cerr << "Options:\n";
        cerr << "       -start <n>     offset to start from (begins with 0) "
                "(only for -all)\n";
        cerr << "       -range <n>     number of order types to process "
                "(only for -all)\n";
        cerr << "       -v             verbose.\n";
        cerr << "       -h             help." << endl;
        exit( ! help);
    }
    int exit_code = 0;
    t_user.start();
    t_real.start();
    if ( all )
	exit_code = count_all( parameters[0], n_points);
    else
	exit_code = count_single( parameters[0]);
    t_user.stop();
    t_real.stop();
    long user_s = long( user_sec + t_user.time());
    long real_s = long( real_sec + t_real.time());
    cerr << "Counting done: " << pseudo.size_of_vertices() << " points, "
	 << user_s/60 << 'm' << user_s%60 << "s user, "
	 << real_s/60 << 'm' << real_s%60 << "s real time."
	 << endl;
    return exit_code;
}
