// ============================================================================
//
// Copyright (c) 2001  Lutz Kettner, see the Copyright file for details
//
// ----------------------------------------------------------------------------
//
// file          : diff_pseudo.C
// revision      : $Revision: 1.1 $
// revision_date : $Date: 2001/09/03 06:12:57 $
// author(s)     : Lutz Kettner
//
// Reads trinum and pseudo counts and compares them. Lists those
// where there are less pseudo than trinum (so far none), and
// lists those where the difference between pseudo and triangulations
// maximizes.
// ============================================================================

#include <CGAL/basic.h>
#include <cstring>

using namespace std;

const char* pseudo_names[11] = {
    "<none>",
    "<none>",
    "<none>",
    "data/pseudo03.b08",
    "data/pseudo04.b08",
    "data/pseudo05.b08",
    "data/pseudo06.b08",
    "data/pseudo07.b16",
    "data/pseudo08.b16",
    "data/pseudo09.b16",
    "data/pseudo10.b32"
};
int pseudo_bytes[11] = { 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 4};
const char* trinum_names[11] = {
    "<none>",
    "<none>",
    "<none>",
    "data/trinum03.b08",
    "data/trinum04.b08",
    "data/trinum05.b08",
    "data/trinum06.b08",
    "data/trinum07.b08",
    "data/trinum08.b08",
    "data/trinum09.b16",
    "data/trinum10.b16"
};
int trinum_bytes[11] = { 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2};


// ----------------------------------------------------------------
int read( istream& in, int k) {
    int i;
    switch (k) {
    case 1:
        {
	    CGAL::UInteger8 j;
	    in.read( & j, 1);
	    i = j;
	    break;
	}
    case 2:
        {
	    CGAL::UInteger16 j;
	    in.read( & j, 2);
	    // swap byte order on big endian machines
	    #ifdef CGAL_BIG_ENDIAN 
	    char* p = reinterpret_cast<char*>( &j);
	    std::swap( p[0], p[1]);
            #endif
	    i = j;
	    break;
	}
    default:
        {
	    CGAL::UInteger32 j;
	    in.read( & j, 4);
	    // swap byte order on big endian machines
	    #ifdef CGAL_BIG_ENDIAN 
	    char* p = reinterpret_cast<char*>( &j);
	    std::swap( p[0], p[3]);
	    std::swap( p[1], p[2]);
            #endif
	    i = j;
	    break;
	}
    }
    return i;
}

// compare: read binary pseudo and trinum data and compare
// ----------------------------------------------------------------
int max_diff = 0;

void compare( int n) {
    CGAL_precondition( n > 2 && n < 11);
    ifstream trinum_in( trinum_names[n]);
    ifstream pseudo_in( pseudo_names[n]);
    long counter      = 0;
    long less_counter = 0;
    while ( true) {
	int trinum = read( trinum_in, trinum_bytes[n]);
	int pseudo = read( pseudo_in, pseudo_bytes[n]);
	if ( ! trinum_in || ! pseudo_in)
	    break;
	if ( pseudo < trinum) {
	    ++ less_counter;
	    if ( less_counter < 10)
		cerr << counter << " # where pseudo < trinum!" << endl;
	} else {
            if ( pseudo - trinum > max_diff)
                max_diff = pseudo - trinum;
        }
	// cout << pseudo << " pseudo ...  " << trinum << " trinum." << endl;
	++counter;
    }
    cerr << counter << " numbers read." << endl;
    cerr << less_counter << " pseudo are less trinum." << endl;
    if ( ! trinum_in.eof() || ! pseudo_in.eof()) {
	cerr << "Error reading files." << endl;
	exit(1);
    }
}

void find_maxdiff( int n) {
    CGAL_precondition( n > 2 && n < 11);
    ifstream trinum_in( trinum_names[n]);
    ifstream pseudo_in( pseudo_names[n]);
    long counter      = 0;
    long max_counter  = 0;
    while ( true) {
	int trinum = read( trinum_in, trinum_bytes[n]);
	int pseudo = read( pseudo_in, pseudo_bytes[n]);
	if ( ! trinum_in || ! pseudo_in)
	    break;
	if ( pseudo - trinum == max_diff) {
	    ++ max_counter;
	    if ( max_counter < 10)
		cerr << counter << " #pseudo - #trinum maximizes!" << endl;
	}
	++counter;
    }
    cerr << max_counter << " times that #pseudo - #trinum miximizes." << endl;
    if ( ! trinum_in.eof() || ! pseudo_in.eof()) {
	cerr << "Error reading files." << endl;
	exit(1);
    }
}


// main function with standard unix commandline arguments
// ----------------------------------------------------------------
int main( int argc, char **argv) {
    const int min_parameters = 1;
    const int max_parameters = 1;
    int n = 0; // number of parameters
    char *parameters[ max_parameters];
    bool help = false;
    for (int i = 1; i < argc; i++) { // check commandline options
        if ( (strcmp( "-h", argv[i]) == 0) || 
	     (strcmp( "-help", argv[i]) == 0))
            help = true;
        else if ( n < max_parameters ) {
            parameters[ n++] = argv[i];
        } else {
	    ++n;
            break;
	}
    }
    if ((n < min_parameters) || (n > max_parameters) || help) {
        if ( ! help)
            cerr << "Error: in parameter list" << endl;
        cerr << "Usage: diff_pseudo [<options>] <n>"<< endl;
	cerr << "    Compares ./data/trinum<n> to ./data/pseudo<n> data files."
             << endl;
        cerr << "    Reports when #triang < #pseudo and reports where the "
                "difference maximizes." << endl;
        cerr << "    All positions in the order type DB are counted beginning "
                "with 0." << endl;
        cerr << "       -h             help." << endl;
        exit( ! help);
    }
    int n_points = atoi( parameters[0]);
    if ( n_points < 3 || n > 10) {
	cerr << "Error: requires an integer in [3,10] as argument." 
	     << endl;
	exit(1);
    }
    compare( n_points);
    find_maxdiff( n_points);
    return 0;
}
