// extract_pseudo_triang.C
// ============================================================================
// revision      : $Revision: 1.1 $
// revision_date : $Date: 2001/09/03 06:12:57 $
// author(s)     : Lutz Kettner  <kettner@inf.ethz.ch>
//
// Extracts one point configuration from a whole file of configurations
// based on Oswin Aichholzer's data base of point order types.
// ============================================================================

#include <CGAL/basic.h>
#include <cstring>

#include <CGAL/Cartesian.h>
#include <CGAL/Point_2.h>
#include <CGAL/Vector_2.h>
#include <CGAL/Aff_transformation_2.h>
#include <PTW/Point_order_types_2.h>

using namespace std;

typedef double                                     NT;
typedef CGAL::Cartesian< NT>                       Kernel;
typedef CGAL::Aff_transformation_2<Kernel>         Affine;
typedef CGAL::Vector_2< Kernel>                    Vector;
typedef CGAL::Point_2< Kernel>                     Point;
typedef PTW::Point_order_types_2< Point*, Kernel>  Point_order_types_2;

// extract: write to cout in ASCII format
// ----------------------------------------------------------------
int extract( const char* filename, int n, long k, bool pretty = false) {
    Affine B( CGAL::SCALING, 1.0 / ((n<9) ? 142.0 : (256.0*142.0)));
    Affine C( CGAL::TRANSLATION, Vector( -0.9, -0.9));
    Affine A = C * B;
    Point points[10];
    Point_order_types_2 order_types( filename, n, points);
    order_types.seek(k);
    if( order_types.read()) {
	if ( pretty ) {
	    CGAL::set_pretty_mode( cout);
            for ( int i = 0; i < n; ++i)
                cout << A(points[i]) << endl;
	} else {
	    CGAL::set_ascii_mode( cout);
            for ( int i = 0; i < n; ++i)
                cout << points[i] << endl;
        }
    } else {
	cerr << "Error: seek/read of point configuration failed." << endl;
	return 1;
    }
    return 0;
}


// main function with standard unix commandline arguments
// ----------------------------------------------------------------
int main( int argc, char **argv) {
    const int min_parameters = 3;
    const int max_parameters = 3;
    int n = 0; // number of parameters
    char *parameters[ max_parameters];
    bool help = false;
    bool pretty  = false;
    for (int i = 1; i < argc; i++) { // check commandline options
	if ( strcmp( "-cgal", argv[i]) == 0)
            pretty = true;
        else if ( (strcmp( "-h", argv[i]) == 0) || 
	     (strcmp( "-help", argv[i]) == 0))
            help = true;
        else if ( n < max_parameters ) {
            parameters[ n++] = argv[i];
        } else {
	    ++n;
            break;
	}
    }
    if ((n < min_parameters) || (n > max_parameters) || help) {
        if ( ! help)
            cerr << "Error: in parameter list" << endl;
        cerr << "Usage: extract_pseudo_triang [<options>] <infile> <n> <k>"
	     << endl;
	cerr << "    Extracts point set <k> from point order type <infile><n>"
	     << endl;
        cerr << "Options:" << endl;
        cerr << "       -cgal          CGAL pretty format for points with "
                          "scaled coordinates." << endl;
        cerr << "       -h             help." << endl;
        exit( ! help);
    }
    int all_n = atoi( parameters[1]);
    if ( all_n < 3 || n > 10) {
	cerr << "Error: second argument needs to be an integer in [3,10]."
	     << endl;
	exit(1);
    }
    long k = atoi( parameters[2]);
    if ( k < 0) {
	cerr << "Error: third argument needs to be an integer >= 0."
	     << endl;
	exit(1);
    }
    return extract( parameters[0], all_n, k, pretty);
}
