// ============================================================================
//
// Copyright (c) 2001  Lutz Kettner, see the Copyright file for details
//
// ----------------------------------------------------------------------------
//
// file          : ptw.C
// revision      : $Revision: 1.2 $
// revision_date : $Date: 2001/09/03 06:12:57 $
// author(s)     : Lutz Kettner
//
// PTW: Pseudo-Triangulation Workbench: 
// main program for interactive exploration
// ============================================================================

#include <CGAL/Cartesian.h>
#include <CGAL/point_generators_2.h> 
#include <CGAL/squared_distance_2.h> 

#include <PTW/PseudoT_default_traits_2.h>
#include <PTW/PseudoT_2.h>
#include <PTW/PseudoT_enumeration_2.h>

#include <CGAL/Timer.h>
#include <CGAL/Real_timer.h>

#include <GL/gl.h>
#include <GL/glut.h>
#include <glui.h>
#include <list>
#include <algorithm>
#include <iostream>
#include <strstream>

using namespace std;

typedef double                                  NT;
typedef CGAL::Cartesian< NT>                    Kernel;
typedef Kernel::Point_2                         Point;
typedef Kernel::Vector_2                        Vector;
typedef Kernel::Segment_2                       Segment;
typedef Kernel::Circle_2                        Circle;
typedef Kernel::Aff_transformation_2            Affine;

typedef PTW::PseudoT_default_traits_2<Kernel>  Traits;
typedef PTW::PseudoT_2<Traits>                 PseudoT;
typedef PTW::PseudoT_enumeration_2<PseudoT>    Enumerator;
typedef PTW::PseudoT_abort_above_degree_n      Abort_above_degree_n;
typedef PTW::PseudoT_enumeration_2<PseudoT, Abort_above_degree_n>
                                                Enumerator_degree;

typedef PseudoT::Vertex_iterator                Vertex_iterator;
typedef PseudoT::Halfedge_iterator              Halfedge_iterator;
typedef PseudoT::Edge_iterator                  Edge_iterator;
typedef PseudoT::Face_iterator                  Face_iterator;
typedef PseudoT::Point_iterator                 Point_iterator;

typedef PseudoT::Vertex_handle                  Vertex_handle;
typedef PseudoT::Halfedge_handle                Halfedge_handle;
typedef PseudoT::Face_handle                    Face_handle;


// =====================================================================
// Model variables
// =====================================================================

int              window_id       = 0;
GLUI*            menu            = 0;
GLUI*            controls        = 0;
GLUI*            stepper         = 0;
GLUI*            status          = 0;
GLUI_StaticText* message         = 0;

GLUI_Listbox*    vertex_color_selector = 0;

const long       ANIM_PROGRESS_STEPS = 100;
const long       RUN_PROGRESS_STEPS  = 5000;
const int        MAX_MESSAGE_BUF     = 1000;
char             message_buf[ MAX_MESSAGE_BUF + 1];
const int        CIRCLE_TESS = 64; // make it smooth, can be a large circle

// Constants
const int   trans_step   = 20;
const float scale_factor = 1.41421356f;

// Input parameters
enum Point_generators {
    random_in_disc = 0,
    regular_on_circle,
    regular_wheel
};
int point_generator = regular_wheel;
int n_points = 8;

// Interaction state
enum Input_state {
    idle_state = 0,
    input_point_state,
    flip_edge_state,
    move_point_state,
    delete_point_state,
    select_face_state
};
enum Algorithm {
    no_algorithm = 0,
    lr_triangulate,
    rl_triangulate,
    lr_pseudo_triangulate,
    rl_pseudo_triangulate,
    greedy_pseudo_triangulate,
    bounded_vertex_triangulate
};
enum Enumeration {
    no_enumeration = 0,
    count_pseudo,
    min_max_degree_enum,
    min_max_6_enum
};
enum Color_vertex_scheme {
    no_vertex_coloring = 0,
    color_reflex,
    color_degree
};
enum Color_edge_scheme {
    no_edge_coloring = 0,
    color_marked,
    color_flippable,
    color_improve_lifting,
    color_improve_angle
};
enum Color_face_scheme {
    no_face_coloring = 0,
    color_circle
};

Input_state       input_state    = idle_state;
Vertex_iterator   selected_vertex;
//Vertex_iterator   selected_vertex2;
//Vertex_iterator   selected_vertex3;
Halfedge_iterator selected_edge;
Face_iterator     selected_face;
Point             last_mouse_position;
int               algorithm      = lr_pseudo_triangulate; // of type Algorithm
bool              dynamic_flag   = false;
Halfedge_handle   flipped_edge;
int               color_vertex_scheme = no_vertex_coloring;
int               color_edge_scheme   = no_edge_coloring;
int               color_face_scheme   = no_face_coloring;

double            min_angle;
Halfedge_handle   min_angle_halfedge;

PseudoT           pseudo;
PseudoT           shadow;

int               enumeration = count_pseudo;
Enumerator        enumerator( pseudo);
Enumerator_degree enumerator_degree( pseudo);
size_t            steps = 0;
size_t            min_max_degree;

// Camera transformation
float scale   = 1.0;
float trans_x = 0.0;
float trans_y = 0.0;

// Appearance in pixel
float point_size      = 4.0;
float line_width      = 3.0;
float node_line_width = 1.0;
float node_diameter   = 8.0;
float threshhold      = 2.0;


// =====================================================================
// Model in world coordinates
// =====================================================================

// search using threshhold and node_diameter as cutoff value
// reports pseudo.vertices_end() if none got found
Vertex_handle find_closest_point( const Point& p) {
    float cutoff = (node_diameter / 2.0 + threshhold) / scale; // in world
    float sqr_cutoff = cutoff * cutoff;
    // reports pseudo.vertices_end() if everything else is larger
    float min_sqr_dist = sqr_cutoff; 
    Vertex_handle result = pseudo.vertices_end();
    for ( Vertex_iterator i = pseudo.vertices_begin(); 
	  i != pseudo.vertices_end(); ++i) {
	float sqr_dist = CGAL::squared_distance(p, i->point());
	if ( sqr_dist <= min_sqr_dist) {
	    min_sqr_dist = sqr_dist;
	    result = i;
	}
    }
    return result;
}

// search using threshhold and line_width as cutoff value
// reports pseudo.halfedge_end() if none got found
Halfedge_handle find_closest_edge( const Point& p) {
    float cutoff = (line_width / 2.0 + threshhold) / scale; // in world
    float sqr_cutoff = cutoff * cutoff;
    // reports pseudo.vertices_end() if everything else is larger
    float min_sqr_dist = sqr_cutoff; 
    Halfedge_iterator result = pseudo.halfedges_end();
    for (Edge_iterator i = pseudo.edges_begin(); i != pseudo.edges_end(); ++i){
	Segment seg( i->vertex()->point(), i->opposite()->vertex()->point());
	float sqr_dist = CGAL::squared_distance(p, seg);
	if ( sqr_dist <= min_sqr_dist) {
	    min_sqr_dist = sqr_dist;
	    result = i;
	}
    }
    return result;
}

// Bettina: vertex degree: 2 = yellow, 3 = green, 4 = red, 5 = pink
// otherwise black

void draw_vertex( Vertex_handle v) {
    const Point& p( v->point());
    if ( selected_vertex == v)
	glColor3f( 1.0, 0.7, 0.2); // orange
//    else if ( selected_vertex2 == v)
//	glColor3f( 0.0, 1.0, 0.0); // green
//    else if ( selected_vertex3 == v)
//	glColor3f( 0.0, 1.0, 0.0); // green
    else {
	glColor3f( 1.0, 0.0, 0.0); // red
	switch ( color_vertex_scheme) {
	case no_vertex_coloring:
	default:
	    break;
	case color_reflex:
	    if ( ! v->is_reflex())
		glColor3f( 1.0, 1.0, 1.0); // white
	    //glColor3f( 1.0, 0.2, 0.8); // pinkish
	    break;
	case color_degree:
	    switch ( v->degree()) {
	    case 2:
		glColor3f( 0.9, 0.9, 0.0); // yellow
		break;
	    case 3:
		glColor3f( 0.0, 0.8, 0.0); // green
		break;
	    case 4:
		glColor3f( 1.0, 0.0, 0.0); // red
		break;
	    case 5:
		glColor3f( 1.0, 0.2, 0.8); // pinkish
		break;
	    default:
		glColor3f( 0.0, 0.0, 0.0); // black
		break;
		
	    }
	    break;
	}
    }
    // 8-gon
    float dd = node_diameter / 2.0 / scale;
    float da = dd / 2.0;
    glBegin( GL_POLYGON);
    glVertex2f( p.x() + dd, p.y() + da);
    glVertex2f( p.x() + dd, p.y() - da);
    glVertex2f( p.x() + da, p.y() - dd);
    glVertex2f( p.x() - da, p.y() - dd);
    glVertex2f( p.x() - dd, p.y() - da);
    glVertex2f( p.x() - dd, p.y() + da);
    glVertex2f( p.x() - da, p.y() + dd);
    glVertex2f( p.x() + da, p.y() + dd);
    glEnd();
    glLineWidth( node_line_width);	      
    glColor3f( 0.0, 0.0, 0.0); // black
    glBegin( GL_LINE_LOOP);
    glVertex2f( p.x() + dd, p.y() + da);
    glVertex2f( p.x() + dd, p.y() - da);
    glVertex2f( p.x() + da, p.y() - dd);
    glVertex2f( p.x() - da, p.y() - dd);
    glVertex2f( p.x() - dd, p.y() - da);
    glVertex2f( p.x() - dd, p.y() + da);
    glVertex2f( p.x() - da, p.y() + dd);
    glVertex2f( p.x() + da, p.y() + dd);
    glEnd();
    glLineWidth( line_width);
}

void draw_edge( Halfedge_handle h) {
    const Point& p = h->vertex()->point();
    const Point& q = h->opposite()->vertex()->point();
    if ( h == selected_edge || h->opposite() == selected_edge)
	glColor3f( 1.0, 0.9, 0.2); // orange
    else if ( h->face()==selected_face || h->opposite()->face()==selected_face)
	glColor3f( 0.8, 0.7, 0.1); // brown
    else {
	glColor3f( 0.2, 0.2, 0.2); // dark gray
	switch ( color_edge_scheme) {
	case no_edge_coloring:
	    break;
	case color_marked:
	    if ( h->is_marked() || h->opposite()->is_marked())
		glColor3f( 0.2, 0.0, 0.6); // darker blue reddish
	    break;
	case color_flippable:
	    if ( pseudo.is_strict_pseudo_flippable( h)) {
		if ( pseudo.pseudo_flip_shares_vertex(( h)))
		    glColor3f( 0.2, 0.0, 0.6); // darker blue reddish
		else
		    glColor3f( 0.5, 0.3, 1.0); // light blue reddish
	    }
	    break;
	case color_improve_lifting:
	    if ( pseudo.is_improving_lifting( h))
		glColor3f( 0.1, 0.8, 0.4); // greenish blue
	    else if ( pseudo.pseudo_flip_shares_vertex(( h)))
		glColor3f( 0.2, 0.0, 0.6); // darker blue reddish
	    break;
	case color_improve_angle:
	    double tmp_angle;
	    if ( pseudo.is_improving_angle( h, tmp_angle)) {
		if ( min_angle_halfedge == h)
		    glColor3f( 0.6, 1.0, 0.8); // light greenish blue
		else
		    glColor3f( 0.1, 0.8, 0.4); // greenish blue
		if ( tmp_angle < min_angle) {
		    min_angle = tmp_angle;
		    min_angle_halfedge = h;
		}
	    }
	    break;
	default:
	    break;
	}
    }
    glBegin( GL_LINES);
    glVertex2f( p.x(), p.y());
    glVertex2f( q.x(), q.y());
    glEnd();
}

void draw_shadow_edge( Halfedge_handle h) {
    const Point& p = h->vertex()->point();
    const Point& q = h->opposite()->vertex()->point();
    glColor3f( 0.8, 0.8, 0.6); // light yellow gray
    glBegin( GL_LINES);
    glVertex2f( p.x(), p.y());
    glVertex2f( q.x(), q.y());
    glEnd();
}

void draw_face( Face_handle f) {
    switch ( color_face_scheme) {
    case no_face_coloring:
    default:
	break;
    case color_circle:
	if ( selected_face != Face_handle() 
	     && selected_face != pseudo.faces_end()
	     && f == selected_face) {
	    Halfedge_handle h1;
	    Halfedge_handle h2;
	    Halfedge_handle h3;
	    if ( f->get_corners( h1, h2, h3)) {
		Circle c( h1->vertex()->point(),
			  h2->vertex()->point(),
			  h3->vertex()->point());
		float r = sqrt( c.squared_radius());
		float phi = (2.0 * M_PI) / CIRCLE_TESS;
		glColor3f( 0.2, 1.0, 0.2); // green
		glBegin( GL_LINE_LOOP);
		for ( int i = 0; i != CIRCLE_TESS; ++i)
		    glVertex2f( c.center().x() + sin(i*phi) * r,
				c.center().y() + cos(i*phi) * r);
		glEnd();
	    }
	}
	break;
    }
}

void draw_world() {
    for ( Face_iterator i = pseudo.faces_begin(); i != pseudo.faces_end(); ++i)
	draw_face( i);
    for ( Edge_iterator i = shadow.edges_begin(); i != shadow.edges_end(); ++i)
	draw_shadow_edge( i);
    min_angle = 7.0; // greater than 2 Pi
    min_angle_halfedge = Halfedge_handle();
    for ( Edge_iterator i = pseudo.edges_begin(); i != pseudo.edges_end(); ++i)
	draw_edge( i);
    if ( min_angle < 7.0)
	draw_edge( min_angle_halfedge);
    for ( Vertex_iterator i = pseudo.vertices_begin(); 
	  i != pseudo.vertices_end(); ++i) {
	draw_vertex( i);
    }
}

void init_model() {
}

// =====================================================================
// Functional Callbacks
// =====================================================================

// forward declarations
void redraw_callback();

void add_point( const Point& p) {
    // check for duplicates
    if ( find( pseudo.points_begin(), pseudo.points_end(), p) 
	 == pseudo.points_end()) {
	pseudo.vertices_push_back( p);
    } else {
	message->set_text( "Error: Point already in list. Not inserted.");
    }
}

void delete_point( Vertex_handle i) {
    pseudo.vertices_erase(i);
}

void flip_edge( Halfedge_handle h) {
    dynamic_flag = false;
    flipped_edge = Halfedge_handle();
    Halfedge_handle g1;
    Halfedge_handle g2;
    if ( pseudo.is_strict_pseudo_flippable( h, g1, g2)) {
	//selected_vertex2 = g1->vertex();
	//selected_vertex3 = g2->vertex();
	flipped_edge = pseudo.pseudo_flip( h, g1, g2);
    } else {
	//selected_vertex2 = g1->vertex();
	//selected_vertex3 = g2->vertex();
	message->set_text( "Error: Edge not strictly flippable in pseudo "
			   "triangulation.");
    }
}

void algorithm_callback( int) {
    flipped_edge = Halfedge_handle();    
    glutSetCursor( GLUT_CURSOR_WAIT);
    switch (algorithm) {
    case lr_triangulate:
	pseudo.triangulate( false);
	break;
    case rl_triangulate:
	pseudo.triangulate( false, true);
	break;
    case lr_pseudo_triangulate:
	pseudo.triangulate( true);
	break;
    case rl_pseudo_triangulate:
	pseudo.triangulate( true, true);
	break;
    case greedy_pseudo_triangulate:
	pseudo.greedy_triangulate();
	break;
    case bounded_vertex_triangulate:
	pseudo.bounded_vertex_triangulate();
	break;
    default:
	break;
    }
    selected_face = pseudo.outer_face();
    glutSetCursor( GLUT_CURSOR_INHERIT);
    glutPostRedisplay();
}

void reset_callback( int) {
    trans_x = 0.0;
    trans_y = 0.0;
    scale   = 1.0;
    glutPostRedisplay();
}

void create_callback( int) {
    // define affine transform such that bbox -1/+1 fits nicely on screen
    int dummy, w, h;
    GLUI_Master.get_viewport_area( &dummy, &dummy, &w, &h);
    float radius = float(h) / 2.1;
    Affine affine = Affine( CGAL::SCALING, 1.0 / scale)
	* Affine( CGAL::TRANSLATION, Vector( w/2.0-trans_x, h/2.0-trans_y))
	* Affine( CGAL::SCALING, radius);
    // Create n points
    size_t n = n_points;
    switch ( Point_generators( point_generator) ) {
    case random_in_disc:
        {
	    CGAL::Random_points_in_disc_2<Point> rnd_points( 1.0);
	    while ( n--)
		add_point( affine( *rnd_points++));
	}
	break;
    case regular_wheel:
	add_point( affine( Point( 0.0, 0.0)));
	--n;
    case regular_on_circle:
        {
	    float phi = (2.0 * M_PI) / n;
	    for ( size_t i = 0; i != n; ++i)
		add_point( affine( Point( sin( i*phi), cos( i*phi))));
	}
	break;
    default:
	break;
    }
    if ( dynamic_flag)
	algorithm_callback(0);
    glutPostRedisplay();
}

void init_callback( int) {
    message->set_text( "Initialized ...");
    dynamic_flag = false;
    steps = 0; // start with 0 since termination counts one too much
    switch ( enumeration) {
    case no_enumeration:
    default:
	message->set_text( "Nothing to do.");
	break;
    case count_pseudo:
	enumerator.init();
	if ( enumerator.is_done()) {
	    std::ostrstream out( message_buf, MAX_MESSAGE_BUF);
	    out << "Done: " << pseudo.size_of_vertices()
		<< " vertices, 1 configuration."
		<< std::ends;
	    message->set_text( message_buf);
	}
	break;
    case min_max_degree_enum:
    case min_max_6_enum:
	vertex_color_selector->set_int_val(color_degree);
	if ( enumeration == min_max_degree_enum)
	    enumerator_degree.abort_branch() = Abort_above_degree_n( 6);
	else
	    enumerator_degree.abort_branch() = Abort_above_degree_n( 5);
	enumerator_degree.init();
	min_max_degree      = pseudo.max_degree();
	shadow              = pseudo;
	if ( enumerator_degree.is_done() || ( enumeration == min_max_6_enum
				              && min_max_degree < 6)) {
	    std::ostrstream out( message_buf, MAX_MESSAGE_BUF);
	    out << "Done: " << pseudo.size_of_vertices()
		<< " vertices, 1 configuration, min-max-degree of "
		<< min_max_degree << "." << std::ends;
	    message->set_text( message_buf);
	}
	break;
    }
    glutPostRedisplay();
}

void step_callback( int) {
    message->set_text( "");
    switch ( enumeration) {
    case no_enumeration:
    default:
	message->set_text( "Nothing to do.");
	break;
    case count_pseudo:
	if ( enumerator.is_done()) {
	    message->set_text( "Counting is ALREADY finished.");
	} else {
	    enumerator.next();
	    ++ steps;
	    if ( enumerator.is_done()) {
		std::ostrstream out( message_buf, MAX_MESSAGE_BUF);
		out << "Done: " << pseudo.size_of_vertices()
		    << " vertices, " << steps << " configurations."
		    << std::ends;
		message->set_text( message_buf);
		std::cerr << message_buf << std::endl;
	    }
	}
	break;
    case min_max_degree_enum:
    case min_max_6_enum:
	if ( enumerator_degree.is_done()) {
	    message->set_text( "Min-max-degree is ALREADY finished.");
	} else {
	    enumerator_degree.next();
	    ++ steps;
	    size_t d = pseudo.bounded_max_degree( min_max_degree);
	    if ( d < min_max_degree) {
		min_max_degree = d;
		shadow         = pseudo;
	    }
	    if ( enumerator_degree.is_done() || ( enumeration == min_max_6_enum
					          && min_max_degree < 6)) {
		std::ostrstream out( message_buf, MAX_MESSAGE_BUF);
		out << "Done: " << pseudo.size_of_vertices()
		    << " vertices, " << steps << " configurations, "
		       "min-max-degree of " << min_max_degree << "." 
		    << std::ends;
		message->set_text( message_buf);
		std::cerr << message_buf << std::endl;
	    }
	}
	break;
    }
    glutPostRedisplay();
}

void run_callback( int animate) {
    message->set_text( (char*)(animate ? "Animate ..." : "Running ..."));
    glutSetCursor( GLUT_CURSOR_WAIT);
    double user_sec = 0.0; // avoid wrap-around of time with big counters
    double real_sec = 0.0; // measuring time in seconds
    CGAL::Timer      t_user;
    CGAL::Real_timer t_real;
    t_user.start();
    t_real.start();
    switch ( enumeration) {
    case no_enumeration:
    default:
	message->set_text( "Nothing to do.");
	glutSetCursor( GLUT_CURSOR_INHERIT);
	glutPostRedisplay();
	return;
    case count_pseudo:
	if ( enumerator.is_done()) {
	    message->set_text( "Counting is ALREADY finished.");
	} else {
	    while ( ! enumerator.is_done()) {
		enumerator.next();
		++ steps;
		if ( (steps % ( animate ? ANIM_PROGRESS_STEPS : 
				          RUN_PROGRESS_STEPS)) == 0) {
		    if ( t_user.time() > 500.0) {
			user_sec += t_user.time();
			t_user.reset(); // avoids wrap-around of timer
		    }
		    if ( t_real.time() > 500.0) {
			real_sec += t_real.time();
			t_real.reset(); // avoids wrap-around of timer
		    }
		    long user_s = long( user_sec + t_user.time());
		    long real_s = long( real_sec + t_real.time());
		    std::ostrstream out( message_buf, MAX_MESSAGE_BUF);
		    out << (animate ? "Animate ..." : "Running ...")
			<< " progress: " << pseudo.size_of_vertices()
			<< " vertices, " << steps << " configurations, "
			<< user_s/60 << 'm' << user_s%60 << "s user, "
			<< real_s/60 << 'm' << real_s%60 << "s real time."
			<< std::ends;
		    message->set_text( message_buf);
		}
		if ( animate) 
		    redraw_callback();
	    }
	}
	break;
    case min_max_degree_enum:
    case min_max_6_enum:
	if ( enumerator_degree.is_done()) {
	    message->set_text( "Min-max-degree is ALREADY finished.");
	} else {
	    while ( ! enumerator_degree.is_done() 
		    && ( enumeration != min_max_6_enum
			 || min_max_degree >= 6)) {
		enumerator_degree.next();
		++ steps;
		size_t d = pseudo.bounded_max_degree( min_max_degree);
		if ( d < min_max_degree) {
		    min_max_degree = d;
		    shadow         = pseudo;
		    redraw_callback();
		}
		if ( (steps % ( animate ? ANIM_PROGRESS_STEPS : 
				          RUN_PROGRESS_STEPS)) == 0) {
		    if ( t_user.time() > 500.0) {
			user_sec += t_user.time();
			t_user.reset(); // avoids wrap-around of timer
		    }
		    if ( t_real.time() > 500.0) {
			real_sec += t_real.time();
			t_real.reset(); // avoids wrap-around of timer
		    }
		    long user_s = long( user_sec + t_user.time());
		    long real_s = long( real_sec + t_real.time());
		    std::ostrstream out( message_buf, MAX_MESSAGE_BUF);
		    out << (animate ? "Animate ..." : "Running ...")
			<< " progress: " << pseudo.size_of_vertices()
			<< " vertices, " << steps << " configurations, "
		           "min-max-degree of " << min_max_degree << ", " 
			<< user_s/60 << 'm' << user_s%60 << "s user, "
			<< real_s/60 << 'm' << real_s%60 << "s real time."
			<< std::ends;
		    message->set_text( message_buf);
		}
		if ( animate) 
		    redraw_callback();
	    }
	}
	break;
    }
    t_user.stop();
    t_real.stop();
    long user_s = long( user_sec + t_user.time());
    long real_s = long( real_sec + t_real.time());
    std::ostrstream out( message_buf, MAX_MESSAGE_BUF);
    out << "Done: " << pseudo.size_of_vertices()
	<< " vertices, " << steps << " configurations, "
	   "min-max-degree of " << min_max_degree << ", " 
	<< user_s/60 << 'm' << user_s%60 << "s user, "
	<< real_s/60 << 'm' << real_s%60 << "s real time."
	<< std::ends;
    message->set_text( message_buf);
    std::cerr << message_buf << std::endl;
    glutSetCursor( GLUT_CURSOR_INHERIT);
    glutPostRedisplay();
}

void clear_callback( int) {
    pseudo.clear();
    reset_callback(0);
}

void clear_edges_callback( int) {
    pseudo.edges_faces_clear();
    glutPostRedisplay();
}

void create_shadow_callback( int) {
    shadow = pseudo;
    glutPostRedisplay();
}

void delete_shadow_callback( int) {
    shadow.clear();
    glutPostRedisplay();
}

void restore_shadow_callback( int) {
    pseudo = shadow;
    shadow.clear();
    glutPostRedisplay();
}

void exit_callback( int return_val) {
    exit( return_val);
}

// =====================================================================
// GUI Callbacks
// =====================================================================

Point mouse_to_world( int x, int y) {
    int dx, dy, w, h;
    GLUI_Master.get_viewport_area( &dx, &dy, &w, &h);
    int height = glutGet( GLenum( GLUT_WINDOW_HEIGHT));
    int top_dy = height - h - dy;
    float world_x = float(x - trans_x) / scale;
    float world_y = float(h - (y - top_dy) - trans_y) / scale;
    return Point( world_x, world_y);
}

void redraw_callback() {
    int x, y, w, h;
    GLUI_Master.get_viewport_area( &x, &y, &w, &h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho( 0, w, 0, h, -1, 1);
    glViewport( x, y, w, h);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef( trans_x, trans_y, 0.0);
    glScalef( scale, scale, scale);
    glClearColor( 0.75, 0.75, 0.75, 0.0); // gray
    glClear(GL_COLOR_BUFFER_BIT);
    glPointSize( point_size);
    glLineWidth( line_width);
    draw_world();
    glFlush();
    glutSwapBuffers();
}

void keyboard_callback( unsigned char key, int, int) {
    message->set_text( "");
    bool key_used = true;
    switch ( key) {
    case 27:
	selected_vertex  = Vertex_handle();
	//selected_vertex2 = Vertex_handle();
	//selected_vertex3 = Vertex_handle();
	selected_edge    = Halfedge_handle();
	selected_face    = Face_handle();
	break;
    case 13:
	algorithm_callback(0);
	break;
    case ' ':
	step_callback(0);
	break;
    case 'c':
    case 'C':
	clear_callback(0);
	break;
    case 'd':
    case 'D':
	dynamic_flag = ! dynamic_flag;
	if ( dynamic_flag)
	    message->set_text( "Dynamic algorithm on.");
	else
	    message->set_text( "Dynamic algorithm off.");
	break;
    case 'f': // flip_back
    case 'F':
	if ( flipped_edge != Halfedge_handle())
	    flip_edge( flipped_edge);
	break;
    case 'i':
    case 'I':
	init_callback(0);
	break;
    case 'p':
    case 'P':
	create_callback(0);
	break;
    case 'Q' - 64:
	exit_callback( 0);
	break;
    case 'r':
    case 'R':
	run_callback(0);
	break;
    case 's':
    case 'S':
	create_shadow_callback(0);
	break;
    case 'v':
	if ( pseudo.is_valid( false, 1))
	    message->set_text( "Is valid pseudo triangulation.");
	else
	    message->set_text( "Is NOT valid pseudo triangulation.");
	break;
    case 'V':
	if ( pseudo.is_valid( true, 1))
	    message->set_text( "Is valid pseudo triangulation.");
	else
	    message->set_text( "Is NOT valid pseudo triangulation.");
	break;
    default:
	key_used = false;
	break;
    }
    if ( key_used) {
	if ( dynamic_flag && key != 13)
	    algorithm_callback(0);
	glutPostRedisplay();
    }
}

void special_key_callback( int key, int x, int y) {
    message->set_text( "");
    (void)x;
    (void)y;
    int dummy, w, h;
    switch (key) {
        case GLUT_KEY_LEFT:
	    trans_x -= trans_step;
	    break;
        case GLUT_KEY_RIGHT:
	    trans_x += trans_step;
	    break;
        case GLUT_KEY_DOWN:
	    trans_y -= trans_step;
	    break;
        case GLUT_KEY_UP:
	    trans_y += trans_step;
	    break;
        case GLUT_KEY_PAGE_UP:
	    GLUI_Master.get_viewport_area( &dummy, &dummy, &w, &h);
	    trans_x = scale_factor * trans_x + (1-scale_factor) *  w / 2.0f;
	    trans_y = scale_factor * trans_y + (1-scale_factor) *  h / 2.0f;
	    //trans_x += (1-scale_factor) * ( w/2.0f - trans_x);
	    scale *= scale_factor;
	    break;
        case GLUT_KEY_PAGE_DOWN:
	    GLUI_Master.get_viewport_area( &dummy, &dummy, &w, &h);
	    scale /= scale_factor;
	    trans_x = (trans_x - (1-scale_factor) * w / 2.0f) / scale_factor;
	    trans_y = (trans_y - (1-scale_factor) * h / 2.0f) / scale_factor;
	    break;
        default:
	    break;
    }
    glutPostRedisplay();
}

void mouse_callback( int button, int action, int x, int y) {
    message->set_text( "");
    Point p = mouse_to_world( x, y);
    if ( action == GLUT_DOWN) {
	Vertex_handle   i = find_closest_point( p);
	if ( button == GLUT_LEFT_BUTTON) {
	    if ( i == pseudo.vertices_end()) {
		Halfedge_handle h = find_closest_edge( p);
		if ( h == pseudo.halfedges_end()) {
		    input_state = select_face_state;
		    selected_edge = pseudo.vertical_ray_shoot(p);
		    if ( selected_edge == Halfedge_handle())
			selected_face = Face_handle();
		    else
			selected_face = selected_edge->face();
		    message->set_text( "Selecting face ....");
		} else {
		    input_state = flip_edge_state;
		    selected_edge = h;
		    selected_vertex = h->vertex();
		    message->set_text( "Flip edge ....");		    
		}
	    } else {
		input_state = move_point_state;
		selected_vertex = i;
		message->set_text( "Move point ....");
	    }
	} else if ( button == GLUT_RIGHT_BUTTON) {
	    if ( i == pseudo.vertices_end()) {
		input_state = input_point_state;
		message->set_text( "Input point ....");
		add_point( Point( p));
		selected_vertex = pseudo.vertices_end();
		--selected_vertex;
	    } else {
		input_state = delete_point_state;
		selected_vertex = i;
		message->set_text( "Delete point ....");
	    }
	}
    }
    if ( action == GLUT_UP) {
	message->set_text( "");
	switch ( input_state) {
	case input_point_state:
	case move_point_state:
	    selected_vertex->point() = selected_vertex->point()
		+ (p - last_mouse_position);
	    break;
	case delete_point_state:
	    {
		CGAL_assertion( selected_vertex != pseudo.vertices_end());
		CGAL_assertion( selected_vertex != Vertex_handle());
		Vertex_handle i = find_closest_point( p);
		if ( i == selected_vertex && 
		     ! (GLUT_ACTIVE_SHIFT & glutGetModifiers()))
		    delete_point( i);
		else
		    message->set_text( "Nothing deleted!");
	    }
	    break;
	case select_face_state:
	    selected_edge = pseudo.vertical_ray_shoot(p);
	    if ( selected_edge == Halfedge_handle())
		selected_face = Face_handle();
	    else
		selected_face = selected_edge->face();
	    break;
	case flip_edge_state:
	    {
		Halfedge_handle h = find_closest_edge( p);
		if ( h == selected_edge && 
		     ! (GLUT_ACTIVE_SHIFT & glutGetModifiers()))
		    flip_edge( h);
		else
		    message->set_text( "No edge flipped!");
		selected_vertex = Vertex_handle();
		selected_edge   = Halfedge_handle();
	    }
	    break;
	default:
	    break;
	}
	//selected_vertex = Vertex_handle();
	//selected_edge   = Halfedge_handle();
	input_state = idle_state;
	if ( dynamic_flag)
	    algorithm_callback(0);
    }
    last_mouse_position = p;
    glutPostRedisplay();
}

void motion_callback( int x, int y) {
    // (x,y) new position relative to upper left corner
    Point p = mouse_to_world( x, y);
    if ( input_state == input_point_state || input_state == move_point_state) {
	selected_vertex->point() = selected_vertex->point() 
	    + (p - last_mouse_position);
	last_mouse_position = p;
	if ( dynamic_flag)
	    algorithm_callback(0);
	glutPostRedisplay();
    } else if ( input_state == select_face_state) {
	selected_edge = pseudo.vertical_ray_shoot(p);
	if ( selected_edge == Halfedge_handle())
	    selected_face = Face_handle();
	else
	    selected_face = selected_edge->face();
	glutPostRedisplay();
    }
}

void reshape_callback( int w, int h) {
    // no rescaling of world
    (void)w;
    (void)h;
}

void create_window() {
    glutInitWindowPosition( 0, 0);
    glutInitWindowSize( 940, 670);
    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE);
    window_id = glutCreateWindow( 
	"Pseudo Triangulation Workbench: $Revision: 1.2 $");
    glutDisplayFunc( redraw_callback);
    glutShowWindow();

    // top control panel -----------------------------------------------
    menu = GLUI_Master.create_glui_subwindow( window_id, GLUI_SUBWINDOW_TOP);
    menu->set_main_gfx_window( window_id);

    GLUI_Listbox* list = menu->add_listbox( "Vertex col", 
					    & color_vertex_scheme);
    vertex_color_selector = list;
    list->add_item( no_vertex_coloring,     "<red>");
    list->add_item( color_reflex,           "Reflex");
    list->add_item( color_degree,           "Degree");
    list->set_int_val(color_vertex_scheme);
    menu->add_column( false);

    list = menu->add_listbox( "Edge", & color_edge_scheme);
    list->add_item( no_edge_coloring,       "<black>");
    list->add_item( color_marked,           "Marked");
    list->add_item( color_flippable,        "Flippable");
    list->add_item( color_improve_lifting,  "Imprv.Lift");
    list->add_item( color_improve_angle,    "Imprv.Angle");
    list->set_int_val(color_edge_scheme);
    menu->add_column( false);

    list = menu->add_listbox( "Face", & color_face_scheme);
    list->add_item( no_face_coloring,       "<none>");
    list->add_item( color_circle,           "Circle");
    list->set_int_val(color_face_scheme);
    menu->add_column( false);

    menu->add_button( "Shadow", 0, create_shadow_callback);
    menu->add_column( false);
    menu->add_button( "Restore Shadow", 0, restore_shadow_callback);
    menu->add_column( false);
    menu->add_button( "Clear Shadow", 0, delete_shadow_callback);
    menu->add_column( false);
    menu->add_button( "Exit", 0, exit_callback);

    // 2nd top control panel--------------------------------------------
    controls = GLUI_Master.create_glui_subwindow( window_id, 
						  GLUI_SUBWINDOW_TOP);
    controls->set_main_gfx_window( window_id);

    list = controls->add_listbox( "Algorithm", & algorithm,
				  0, algorithm_callback);
    list->add_item( no_algorithm,               "<none>");
    list->add_item( lr_triangulate,             "LR Triang");
    list->add_item( rl_triangulate,             "RL Triang");
    list->add_item( lr_pseudo_triangulate,      "LR Pseudo");
    list->add_item( rl_pseudo_triangulate,      "RL Pseudo");
    list->add_item( greedy_pseudo_triangulate,  "Greedy Pseudo");
    //list->add_item( bounded_vertex_triangulate, "Bounded Vertex");
    list->set_int_val( algorithm);
    controls->add_column( false);
    GLUI_Spinner* spinner = controls->add_spinner( "#Pts", 
						   GLUI_SPINNER_INT,
						   & n_points);
    spinner->set_int_limits( 0, 1000);
    //spinner->set_speed( 0.1);
    controls->add_column( false);
    list = controls->add_listbox( "Generator", & point_generator);
    list->add_item( random_in_disc,       "Rnd in disc");
    list->add_item( regular_on_circle,    "Reg on circle");
    list->add_item( regular_wheel,        "Reg wheel");
    list->set_int_val( random_in_disc);
    controls->add_column( false);

    controls->add_button( "Create", 0, create_callback);
    controls->add_column( false);
    controls->add_button( "Clear Edges", 0, clear_edges_callback);
    controls->add_column( false);
    controls->add_button( "Clear", 0, clear_callback);
    controls->add_column( false);
    controls->add_button( "Reset View", 0, reset_callback);

    // 3rd top control panel--------------------------------------------
    stepper = GLUI_Master.create_glui_subwindow( window_id, 
						 GLUI_SUBWINDOW_TOP);
    stepper->set_main_gfx_window( window_id);

    list = stepper->add_listbox( "Enum", & enumeration);
    list->add_item( no_enumeration,      "<none>");
    list->add_item( count_pseudo,        "Count Pseudo");
    list->add_item( min_max_degree_enum, "Min-max-degree");
    list->add_item( min_max_6_enum,      "Min-max-6");
    list->set_int_val( enumeration);
    stepper->add_column( false);

    stepper->add_button( "Init", 0, init_callback);
    stepper->add_column( false);
    stepper->add_button( "Step", 0, step_callback);
    stepper->add_column( false);
    stepper->add_button( "Animate", 1, run_callback);
    stepper->add_column( false);
    stepper->add_button( "Run", 0, run_callback);

    // lower control panel ---------------------------------------------
    status = GLUI_Master.create_glui_subwindow( window_id, 
						GLUI_SUBWINDOW_BOTTOM);
    status->set_main_gfx_window( window_id);
    message = status->add_statictext( 
	"Left-click moves points, flips edges, or selects faces.   "
	"Right-click creates or deletes points.");

    // general ---------------------------------------------------------
    GLUI_Master.set_glutKeyboardFunc( keyboard_callback);
    GLUI_Master.set_glutSpecialFunc(  special_key_callback);
    GLUI_Master.set_glutMouseFunc(    mouse_callback);
    GLUI_Master.set_glutReshapeFunc(  reshape_callback);
    glutMotionFunc( motion_callback);
    // glutPassiveMotionFunc( motion_callback);
    glutSetWindow(window_id);
}


// =====================================================================
// main
// =====================================================================

int main( int argc, char* argv[]) {
    glutInit(&argc, argv);
    init_model();
    create_window();
    glutMainLoop();
    return 0;
}

// EOF //


